/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"output.h"
#include"twister.h"
#include"extstruct.h"


/******************************************************************************/
/*                              ExtendStructure                               */
/******************************************************************************/
void ExtendStructure(NETWORK nt)
{
  // variables for DEBUGGING
  //int i, j;

  // WARNING:  hard-coded for one hybrid subtree only!
  
  /************************************************************************/
  /*  this is where we add in the hybrid parents including setting their  */
  /*  sequences -- the outgroup is also removed and adjusted if requested */
  /************************************************************************/

  // before anything else, make the node reticualte now
  nt->hybrid_subtree_root->reticulate = 1;
  nt->num_unique_reticulate_nodes++;

  // clean up the outgroup as it not really part of this subtree
  UnhookHybridSubtreeOutgroup(nt);

  // now make the two parental nodes and get them hooked in properly
  CreateHybridParentalNodes(nt);

  // next we need to split the sequences between the two parents
  AssignDipHybridParentalSeqs(nt);

  /*********************/
  /*      DEBUGGING    */
  /*********************/
  //printf("NR DEBUG: hybrid seqs are:\n");
  //for (i = 0; i < num_hom_chromos; i++)
  //{
    //for (j = 0; j < num_strands_per_hom_chrom; j++)
    //{
      //printf("%s\n", nt->hybrid_subtree_root->hom_chromos[i]->strands[j]->sequence);
    //}
  //}
  //printf("NR DEBUG: Parent 1 seqs are:\n");
  //for (i = 0; i < num_hom_chromos; i++)
  //{
    //for (j = 0; j < num_strands_per_hom_chrom; j++)
    //{
      //printf("%s\n", nt->hybrid_subtree_root->parent1_node->hom_chromos[i]->strands[j]->sequence);
    //}
  //}
  //printf("NR DEBUG: Parent 2 seqs are:\n");
 // for (i = 0; i < num_hom_chromos; i++)
 // {
  //  for (j = 0; j < num_strands_per_hom_chrom; j++)
   // {
    //  printf("%s\n", nt->hybrid_subtree_root->parent2_node->hom_chromos[i]->strands[j]->sequence);
   // }
 // }

  /***********************/
  /*    END DEBUGGING    */
  /***********************/

  return;
} 

/******************************************************************************/
/*                        UnhookHybridSubtreeOutgroup                         */
/******************************************************************************/
void UnhookHybridSubtreeOutgroup(NETWORK nt)
{
  NODE og;
  NODE new_root;
  NODE old_root;
  int na_index;
  int hchromo_i;
  int strand_i;

  og = nt->outgroups_for_subtrees[0];

  // first deal with the direction from parent to children
  if (nt->hybrid_subtree_root->first_child_node == og)
  {
    if (og->next_sib_node == NULL)
    {
      printf("NR ERROR:  Something went awry with the hybrid subtree topology as the outgroup is the only child of the root -- program aborted\n");
      exit(0);
    }
    else
    {
      nt->hybrid_subtree_root->first_child_node = og->next_sib_node; 
      og->next_sib_node->prev_sib_node = NULL;
    }
  }
  else // og is a sibling
  {
    if (og->next_sib_node != NULL)
    {
      og->next_sib_node->prev_sib_node = og->prev_sib_node;
      og->prev_sib_node->next_sib_node = og->next_sib_node;
    }
    else
    {
      og->prev_sib_node->next_sib_node = NULL;
    }
  }

  // now reset outgroup values
  og->parent1_node = NULL;
  og->next_sib_node = NULL;
  og->prev_sib_node = NULL;

  /**********************************************************************/
  /* Now we need to see if the hybrid's subtree root has more than one  */
  /* child remaining, if it does, we may want to adjust the toplogy so  */
  /* that there is only one child from the hybrid root first and then   */
  /* the topology continues as reconstructed.  Namely if we have        */
  /* something like the left, we want something like the right:         */
  /*                                                                    */
  /*                          H                                         */
  /*     H                    |                                         */
  /*   / | \                  |                                         */
  /*  /  |  \               / | \                                       */
  /*                       /  |  \                                      */
  /*                                                                    */
  /* If there is only one from the root, already, nothing to do.        */
  /*                                                                    */
  /* NOTE: all this work only gets done if the outgroup adjustment      */
  /*       flag is set to true (1) which is the default...              */
  /**********************************************************************/

  if ( (outgroup_adjustment == 1) &&
       (nt->hybrid_subtree_root->first_child_node->next_sib_node != NULL) )
  {
    // make a new node that will be the new root of the subtree with
    // a single child -- the old root of the subtree
    na_index = nt->node_id_ctr++;
    new_root = NewNode(na_index);

    // up the counter
    nt->num_unique_nodes = nt->num_unique_nodes + 1;

    old_root = nt->hybrid_subtree_root;
    old_root->parent1_node = new_root;
    old_root->reticulate = 0;
    new_root->first_child_node = old_root;
    new_root->reticulate = 1;

    // the new root needs to have the exact same sequences as the old root
    // so copy them in here...
    for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
    {
      for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
      {
        strcpy(new_root->hom_chromos[hchromo_i]->strands[strand_i]->sequence,
	       old_root->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
      }
    }

    // put the new node in the array and set it up as the rightful root
    // of the hybrid subtree 
    nt->NodeArray[na_index] = new_root;
    nt->hybrid_subtree_root = new_root;
    nt->subtree_roots[0] = new_root;
    nt->root = new_root; // at this point the root of the network still 
                         // needs to point to this root as well

    // DEBUGGING
    printf("NR DEBUG:  Hybrid subtree root WAS adjusted due to outgroup polytomy\n");
  }

  return;
}

/******************************************************************************/
/*                       CreateHybridParentalNodes                            */
/******************************************************************************/
void CreateHybridParentalNodes(NETWORK nt)
{
  int na_index;
  NODE p1;
  NODE p2;

  // make the first parent
  na_index = nt->node_id_ctr++;

  p1 = NewNode(na_index); 
  nt->hybrid_subtree_root->parent1_node = p1;
  p1->first_child_node = nt->hybrid_subtree_root;
  nt->NodeArray[na_index] = p1;

  // make the second parent
  na_index = nt->node_id_ctr++;

  p2 = NewNode(na_index); 
  nt->hybrid_subtree_root->parent2_node = p2;
  p2->first_child_node = nt->hybrid_subtree_root;
  nt->NodeArray[na_index] = p2;

  // make sure we assign the two branch lengths from each parent into the
  // hybrid as 0.0 since there is no evolution occurring there
  nt->hybrid_subtree_root->brlen1 = 0.0;
  nt->hybrid_subtree_root->brlen2 = 0.0;

  // up the node counter
  nt->num_unique_nodes = nt->num_unique_nodes + 2;

  // DEBUGGING
  //printf("NR DEBUG: the internal ids for the two parents of the hybrid are %d  %d and the hybrid itself is %d\n", nt->hybrid_subtree_root->parent1_node->internal_node_id, nt->hybrid_subtree_root->parent2_node->internal_node_id, nt->hybrid_subtree_root->internal_node_id);

  //printf("NR DEBUG: the hybrid status of the hybrid node is %d\n", nt->hybrid_subtree_root->reticulate);

  return;
}

/******************************************************************************/
/*                       AssignDipParentalHybridSeqs                          */
/******************************************************************************/
void AssignDipHybridParentalSeqs(NETWORK nt)
{
  NODE p1;
  NODE p2;
  int hchromo_i, strand_i, i;
  int p1_spot_ctr, p2_spot_ctr, half_spots_to_fill;
  int max_spots, rand_spot, spot_found, rand_try_ctr, boundary, adj_spot;
  int used_indices[2 * MAX_NUM_STRANDS_PER_HC];

  /************************************************************************/
  /* This is where we split up the hybrid (really the above hybrid node's */
  /* sequences to the two parents -- which is the reverse of what happens */
  /* with diploid hybrids -- the unknowns are filled with ?s as that      */
  /* is what phylip will need.                                            */
  /************************************************************************/
  
  p1 = nt->hybrid_subtree_root->parent1_node;
  p2 = nt->hybrid_subtree_root->parent2_node;

  // first fill both parental seqs with ?s
  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      strcpy(p1->hom_chromos[hchromo_i]->strands[strand_i]->sequence, "\0");
      strcpy(p2->hom_chromos[hchromo_i]->strands[strand_i]->sequence, "\0");

      for (i = 0; i < sequence_length; i++)
      {
         strcat(p1->hom_chromos[hchromo_i]->strands[strand_i]->sequence, "?");
         strcat(p2->hom_chromos[hchromo_i]->strands[strand_i]->sequence, "?");
      }
    }
  }

  // now we need to do lots of work to randomly split up the sequences

  max_spots = 2 * num_strands_per_hom_chrom;
  half_spots_to_fill = num_strands_per_hom_chrom / 2;
  boundary = (max_spots / 2) - 1; // adj for 0 start in array

  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    // first initialize our used_indices array which is all
    // strand spots for this homologous chromosome for two parents
    for (i = 0; i < max_spots; i++)
    {
      used_indices[i] = 0;
    }

    // initialize the spot counters as well
    p1_spot_ctr = 0;
    p2_spot_ctr = 0;

    // now work with the individual strand sequences
    // but don't forget it has to be half and half split
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      // initialize our random try counter so we don't end up with
      // infinite loop possible problems
      rand_try_ctr = 0;

      // gives us an index into which strand spot will be occcupied
      // by this sequence, the first half are for p1 and the second
      // half are for p2
      rand_spot = (int) (genrand_real2() * max_spots);

      spot_found = 0;

      // now we need to see if the rand spot is already occupied or
      // not, if not and that parent still needs sequences, go ahead 
      // and assign seq to that spot, otherwise  must reselect a 
      // rand_spot again
      while( (spot_found == 0) && (rand_try_ctr < 1000000) )
      {
        if ( (used_indices[rand_spot] == 0) &&
	     (rand_spot <= boundary) &&
	     (p1_spot_ctr < half_spots_to_fill) )
	{
	  spot_found = 1;
	  p1_spot_ctr++;
	}
	else if ( (used_indices[rand_spot] == 0) &&
	          (rand_spot > boundary)  && 
		  (p2_spot_ctr < half_spots_to_fill) )
	{
	    spot_found = 1;
	    p2_spot_ctr++;
	}
	else
	{
	  rand_spot = (int) (genrand_real2() * max_spots);
	  rand_try_ctr++;
	}
      }

      if (spot_found == 0)
      {
        printf("NR ERROR:  exceeded max number of tries for assigning diploid hybrid seqs to parents (hchromo %d strand %i) -- program aborted\n", hchromo_i, strand_i);
	exit(0);
      }
      else // time to give the sequence to its home, record the used index
           // and also note that the sequence is "fixed" for future processing
      {
        used_indices[rand_spot] = 1;

        if (rand_spot <= boundary) // will be a p1 sequence
	{
	  strcpy(p1->hom_chromos[hchromo_i]->strands[rand_spot]->sequence,
	       nt->hybrid_subtree_root->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

	  p1->hom_chromos[hchromo_i]->strands[rand_spot]->fixed_seq = 1;
	}
	else // will be a p2 sequence -- adjust index
	{
	  adj_spot = rand_spot - num_strands_per_hom_chrom;

	  strcpy(p2->hom_chromos[hchromo_i]->strands[adj_spot]->sequence, 
	      nt->hybrid_subtree_root->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

	  p2->hom_chromos[hchromo_i]->strands[adj_spot]->fixed_seq = 1;
	}
      }
    } // next strand
  }

  return; 
}
