/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"output.h"
#include"hybsub.h"
#include"twister.h"
#include"finnet.h"


/******************************************************************************/
/*                               FinishNetwork                                */
/******************************************************************************/
void FinishNetwork(NETWORK nt)
{
  // now we need to assign sequence information to all nodes
  // that do not have it already -- those that have it already
  // are the hybrid and its subtree (done in reconsub.c) and
  // the extant leaves -- the parental hybrids only have
  // half their sequences and will thus need assignment of
  // the others
  AssignRemSeqs(nt);

  return;
} 

/******************************************************************************/
/*                              AssignRemSeqs                                 */
/******************************************************************************/
void AssignRemSeqs(NETWORK nt)
{
  int node_i, base_i;
  int hchromo_i, strand_i;

  // DEBUGGING
  //int base_j;

  // This routine is similar to the AssignSubtreeSeqs found
  // in hybsub.c, but has been modified for this stage.

  // General idea is to work your way up from the "leaves"
  // used in the reconstruction (the hybrid node and
  // its offspring already have all their seqs) to give
  // the options and then from the root down during the
  // downward phase assign actual bases.  This approach
  // is known as the Fitch small parsimony algorithm.
  // This is a VERY simplistic approach, but should be 
  // easy to follow hopefully.

  // The nodes that are like the "leaves" for this stage
  // are all the extant taxa that were not impacted by the
  // hybrid, the parental hybrid nodes, and the outgroup.
  // Mark them so we can easily identify them for populating
  // purposes.

  ResetMarkersAll(nt);

  /*************/
  /*   mark    */
  /*************/
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if ( ((nt->NodeArray[node_i]->extant_leaf) &&
          (nt->NodeArray[node_i]->impacted_by_hybrid[0] == -1)) ||
	 ((nt->NodeArray[node_i]->first_child_node != NULL) &&
          (nt->NodeArray[node_i]->first_child_node->reticulate == 1)) )
    {
      nt->NodeArray[node_i]->marker = 1;
    }
  }

  /*****************************************************/
  /*    first populate the appropriate option arrays   */
  /*****************************************************/
  for(node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->marker == 1)
    {
      PopulateExtantBaseValues(nt->NodeArray[node_i]);
    }
  }

  /****************************/
  /*    upward option phase   */
  /****************************/

  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      for (base_i = 0; base_i < sequence_length; base_i++)
      {
        last_hyb_parent_hit = 0;
        hyb_parent_hit_ctr = 0;

	// we do a postorder traversal from the root to get things 
	// percolated up from the leaves to the root so we have the
	// option array set up correctly
        UpwardOptionRound(nt->root, base_i, hchromo_i, strand_i);
      }
    } 
  } 

  /*******************************/
  /*         DEBUGGING           */
  /*******************************/
  //printf("NR DEBUG:  NODE OPTIONS -- CONNECT NETWORK\n");
  //printf("(note -- all nodes will be listed here, including those of subtree previously processed, but those nodes will have their option array already set to their sequence (i.e. no option))\n\n");

  //printf("internal id for root is %d\n", nt->root->internal_node_id);

  //for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  //{
    //printf("node %d (external id) %d (internal id)\n", nt->NodeArray[node_i]->external_node_id, nt->NodeArray[node_i]->internal_node_id);

    //for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
    //{
      //for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
      //{
        //for (base_j = 0; base_j < 4; base_j++)
        //{
          //for (base_i = 0; base_i < sequence_length; base_i++)
          //{
            //printf("%d ", nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_j][base_i]);
          //}
          //printf("\n");
        //}
        //printf("\n");
      //}
      //printf("\n");
    //}
   // printf("\n");
  //}
  /*******************************/
  /*        END DEBUGGING        */
  /*******************************/

  /*****************************************************************/
  /*                      now do the downward work                 */
  /*****************************************************************/

  // NOTE:  we skip over a reticulate node and its descendants
  //        and before setting seqs we check the fixed_seq 
  //        parameter which is set to 1 (positve) only for 
  //        those parental hybrid nodes' sequences that were 
  //        "brought up" from the hybrid node itself and
  //        extant taxa

  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      for (base_i = 0; base_i < sequence_length; base_i++)
      {
	// now do the assignment for this base
        DownwardAssignRound(nt->root, base_i, hchromo_i, strand_i);
      }
    }
  }

  /*******************************/
  /*         DEBUGGING           */
  /*******************************/
  // print out the options for all the nodes
//
//  printf("NR DEBUG: NODE ASSIGNMENTS\n\n");
//
//  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
//  {
//    printf("node %d (external id) %d (internal id)   reticulate status %d \n", nt->NodeArray[node_i]->external_node_id, nt->NodeArray[node_i]->internal_node_id, nt->NodeArray[node_i]->reticulate);
//
//    for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
//    {
//      for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
//      {
//        for (base_j = 0; base_j < 4; base_j++)
//        {
//          for (base_i = 0; base_i < sequence_length; base_i++)
//          {
//            printf("%d ", nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_j][base_i]);
//          }
//          printf("\n");
//        }
//        printf("\n");
//      }
//      printf("\n");
//    }
//    printf("\n");
//  }
  /*******************************/
  /*        END DEBUGGING        */
  /*******************************/

  return;
}
