/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"phylipinter.h"
#include"randipinter.h"
#include"output.h"
#include"twister.h"
#include"hybsub.h"


/******************************************************************************/
/*                               ReconSubtree                                 */
/******************************************************************************/
void ReconHybridSubtree(NETWORK nt)
{
  int subtree_id;
  FILE *pipe_ptr;

  // WARNING:  all of this routine is based upon ONE hybrid and its parents
  subtree_id = 0;

  PrepareAndWriteInputFile(nt, subtree_id);

  // now that we have our input file made, we need to run a PHYLIP
  // routine (depends on how many impacted taxa we are dealing with 
  // for this subtree -- thresh or less, dnapenny, otherwise neighbor)
  // (the outgroup acts as one more added to the inputted threshold)

  // do reconstruction for this phase based upon what type requested
  if (recon_type == 0) // PHYLIP
  {
    if (num_impacted_by_hybrid <= mp_phylip_thresh)
    {
      ExecuteDnaPenny(nt, subtree_id);

      if ((pipe_ptr = popen("mv outfile outfile_dna_penny \n mv outtree outtree_dna_penny\n", "r")) != NULL)
      {
        // close the pipe
        (void) pclose(pipe_ptr);
      }
      else
      {
        printf("NR ERROR:  Unexpected issue when moving outtree or outfile after running PHYLIP's dnapenny to create the hybrid subtree -- program aborted\n");
        exit(0);
      }

      ExecuteConsense(nt, subtree_id); // doing Majority-Rule extended
    }
    else // more than 9 impacted by hybrid taxa
    {
      ExecuteDnaDist(nt);

      if ((pipe_ptr = popen("mv outfile outfile_dna_dist \n", "r")) != NULL)
      {
        // close the pipe
        (void) pclose(pipe_ptr);
      }
      else
      {
        printf("NR ERROR:  Unexpected issue when moving outfile after running PHYLIP's dnadist during hybrid subtree work -- program aborted\n");
        exit(0);
      }

      ExecuteNeighbor(nt);
    }
  }
  else if (recon_type == 1) // RANDIP
  {
    ExecuteRandIP(nt, subtree_id);
  }
  else
  {
    printf("NR ERROR:  Invalid reconstruction algorithm chosen for hybrid subtree -- program aborted\n");
    exit(0);
  }

  // now we need to rename the final output files
  if ((pipe_ptr = popen("mv outfile outfile_hybrid_subtree \n mv outtree outtree_hybrid_subtree\n", "r")) != NULL)
  {
    // close the pipe
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NR ERROR:  Unexpected issue when moving outtree or outfile after running PHYLIP's consense/neighbor (or RANDIP) to create the hybrid subtree -- program aborted\n");
    exit(0);
  }

  // now reading the results of above need to Create the Subtree Structure
  ReadHybridSubtreeTopology(nt);
 
  // now we need to assign sequences to all the internal nodes, for
  // this subtree using the algorithm known for small parsimony
  // problem attributed to Fitch
  AssignHybridSubtreeSeqs(nt, subtree_id);

  // before we leave here, get the pointer to the subtree root set
  // to the root, which currently holds it
  nt->hybrid_subtree_root = nt->root;
  nt->subtree_roots[0] = nt->root;

} 

/******************************************************************************/
/*                        PrepareAndWriteInputFile                            */
/******************************************************************************/
void PrepareAndWriteInputFile(NETWORK nt, int subtree_id)
{
  int node_i;
  NODE og;


  num_impacted_by_hybrid = CountExtantImpactedByHybrid(nt);

  if  (num_impacted_by_hybrid == 0)
  {
    printf("NR ERROR: there are no extant taxa identified as being impacted by the hbyrid -- this is not allowed -- program aborted\n");
    exit(0);
  }

  // find the outgroup for the hybrid subtree so we can send it to
  // the input writing routine
  og = NULL;

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i] == nt->outgroups_for_subtrees[subtree_id])
      og = nt->NodeArray[node_i];
  }

  if (og == NULL)
  {
    printf("NR ERROR:  Could not find outgroup for hybrid subtree -- program aborted\n");
    exit(0);
  }

  ResetWriteMarkers(nt);

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->impacted_by_hybrid[0] != -1)
    {
      nt->NodeArray[node_i]->marked_to_write = 1;
    }
  }

  // now we need to write the input file for PHYLIP
  WriteInputFile(nt, num_impacted_by_hybrid, og, subtree_id);

  return;
   
}

/******************************************************************************/
/*                        ReadHybridSubtreeTopology                           */
/******************************************************************************/
void ReadHybridSubtreeTopology(NETWORK nt)
{
  // first we need to open the outtree file
  fp_temp_in = fopen("outtree_hybrid_subtree","r");

  if (fp_temp_in == NULL)
  {
    printf("NR ERROR:  Cannot open the temp input file for reading PHYLIP's hybrid subtree results (outtree_hybrid_subtree) -- program aborted\n");
    exit(0);
  }

  // now we need to do the parsing
  ParseNewick(nt);

  // finally close the file
  fclose(fp_temp_in);

  return;
}

/******************************************************************************/
/*                         AssignHybridSubtreeSeqs                            */
/******************************************************************************/
void AssignHybridSubtreeSeqs(NETWORK nt, int subtree_id)
{
  int og, node_i, base_i;
  int hchromo_i, strand_i;

  // DEBUGGING
  //int base_j;

  // This routine assigns sequences for the subtree using the idea
  // of the small parsimony problem.  Speicifically work your way
  // from the leaves to the root listing what bases are options
  // for each index (must be done as a post-order traversal from
  // the root), and then starting at the root, work your
  // way down assigning what the bases are for each index.  This
  // is commonly known as the fitch small parsimony problem.  This
  // is a VERY simplistic implementation, but should be easy
  // enough to follow the logic.


  // WARNING :  this logic works for TREEs not NETWORKs -- we
  //            should only be dealing with a subTREE at this point
  if (!IsTree(nt))
  {
    printf("NR ERROR:  AssignHybridSubtreeSeqs routine has been called for a non-tree structure which is incorrect logic -- program aborted\n");
    exit(0);
  }
  
  /************************************************************/
  /*    first populate the appropriate extant option arrays   */
  /************************************************************/
  og = nt->outgroup_ids_for_subtrees[subtree_id];

  for(node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if ( (nt->NodeArray[node_i]->extant_leaf == 1) &&
         ((nt->NodeArray[node_i]->impacted_by_hybrid[0] == subtree_id) || 
	  (nt->NodeArray[node_i]->external_node_id == og)) )
    {
      PopulateExtantBaseValues(nt->NodeArray[node_i]);
    }
  }

  /************************************************************/
  /*            now figure out the options                    */
  /************************************************************/
  // WARNING:  remember this version only does one subtree 
  ResetMarkersAll(nt);

  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      for (base_i = 0; base_i < sequence_length; base_i++)
      {
        last_hyb_parent_hit = 0;
	hyb_parent_hit_ctr = 0;

	// we do a postorder traversal from the root to get things 
	// percolated up from the leaves to the root so we have the
	// option array set up correctly
        UpwardOptionRound(nt->root, base_i, hchromo_i, strand_i);

      }
    } 
  } 

  /*******************************/
  /*         DEBUGGING           */
  /*******************************/
//  printf("NR DEBUG: NODE OPTIONS\n\n");
//
//  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
//  {
//    printf("node %d (external id) %d (internal id)\n", nt->NodeArray[node_i]->external_node_id, nt->NodeArray[node_i]->internal_node_id);
//
//    for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
//    {
//      for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
//      {
//        for (base_j = 0; base_j < 4; base_j++)
//        {
//          for (base_i = 0; base_i < sequence_length; base_i++)
//          {
//            printf("%d ", nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_j][base_i]);
//          }
//          printf("\n");
//        }
//        printf("\n");
//      }
//      printf("\n");
//    }
//    printf("\n");
//  }
  /*******************************/
  /*        END DEBUGGING        */
  /*******************************/


  /*********************************************/
  /*        now do the downward work           */
  /*********************************************/
  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      for (base_i = 0; base_i < sequence_length; base_i++)
      {
        DownwardAssignRound(nt->root, base_i, hchromo_i, strand_i);
      }
    }
  }

  /*******************************/
  /*         DEBUGGING           */
  /*******************************/
  // print out the options for all the nodes
//
//  printf("NR DEBUG: NODE ASSIGNMENTS\n\n");
//
//  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
//  {
//    printf("node %d (external id) %d (internal id)\n", nt->NodeArray[node_i]->external_node_id, nt->NodeArray[node_i]->internal_node_id);
//
//    for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
//    {
//      for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
//      {
//        for (base_j = 0; base_j < 4; base_j++)
//        {
//          for (base_i = 0; base_i < sequence_length; base_i++)
//          {
//            printf("%d ", nt->NodeArray[node_i]->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_j][base_i]);
//          }
//          printf("\n");
//        }
//        printf("\n");
//      }
//      printf("\n");
//    }
//    printf("\n");
//  }
  /*******************************/
  /*        END DEBUGGING        */
  /*******************************/

  return;
}

