/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <ctype.h>
#include <curses.h>
#include "globals.h"
#include "node.h"
#include "network.h"
#include "output.h"
#include "input.h"

/*****************************************************************************/
/*                         Read the input file!                              */
/*****************************************************************************/
void ReadInput(NETWORK nt) 
{
  char str[MAX_LINE_LENGTH];
  char * eol_ptr;

  eol_ptr = NULL;

  // set all default values -- alphabetical order
  dna_dist_model = 0;  // if not specified, default jukes cantor used
  dna_dist_model_yn = 0; // 0 not specified, 1 it is specified
  extant_taxa_read = 0;
  mp_phylip_thresh = MP_PHYLIP_THRESH_DEFAULT; // threshold for dnapenny
  num_strands_per_hom_chrom = 0;
  num_hom_chromos = 0;
  num_extant_total = 0;
  outgroup_adjustment = 1;
  outgroup_taxon_id = -1;
  parental_offspring_id_method = 0; // default approach
  recon_type = 0; // default is to use phylip
  sequence_length = 0;
  subtree_listing_ctr = 0;


  while (fscanf(fp_in, "%s", str) != EOF)
  {
    if (strncmp(str,"#",1) == 0) {
      if (strchr(str, '\n') == NULL)
      {
        // no end of line found so do getline and ignore rest of line
	fgets(str, MAX_LINE_LENGTH, fp_in);
      }
      else
      {
        // end of line was found, so do nothing but proceed with processing
	;
      }
    } else if (strcmp(str, "extant_taxa_listing") == 0) {
      ReadExtantTaxa(nt);
      extant_taxa_read = 1;
    } else if (strcmp(str, "begin_subtree_listing") == 0) {
      ReadSubtreeInfo(nt);
    } else if (strcmp(str, "mp_phylip_thresh") == 0) {
      fscanf(fp_in, "%d", &mp_phylip_thresh);
    } else if (strcmp(str, "num_strands_per_homologous_chromosome") == 0) {
      fscanf(fp_in, "%d", &num_strands_per_hom_chrom);
    } else if (strcmp(str, "num_homologous_chromosomes") == 0) {
      fscanf(fp_in, "%d", &num_hom_chromos);
    } else if (strcmp(str, "num_extant_total") == 0) {
      fscanf(fp_in, "%d", &num_extant_total);
    } else if (strcmp(str, "outgroup_adjustment") == 0) {
      fscanf(fp_in, "%d", &outgroup_adjustment);
      nt->outgroup_external_id = outgroup_taxon_id;
    } else if (strcmp(str, "outgroup_taxon_id") == 0) {
      fscanf(fp_in, "%d", &outgroup_taxon_id);
      nt->outgroup_external_id = outgroup_taxon_id;
    } else if (strcmp(str, "sequence_length") == 0) {
      fscanf(fp_in, "%d", &sequence_length);
    } else if (strcmp(str, "recon_type_PHYLIP") == 0) {
      recon_type = 0;
    } else if (strcmp(str, "recon_type_RANDIP") == 0) {
      recon_type = 1;
    } else if (strcmp(str, "dna_dist_model") == 0) {
      dna_dist_model_yn = 1;
      ReadDnaDistModel();
    } else {
      printf("NR ERROR:  Unknown parameter in input file: %s\n", str);
      exit(0);
    }
  }

  CheckPostEssentials(nt);

}

/*****************************************************************************/
/*                   Read (& Create) Extant Taxa                             */
/*****************************************************************************/
void ReadExtantTaxa(NETWORK nt)
{
  int ex_id;
  int extant_i, seq_i, hchromo_i, strand_i;
  int na_index;
  int nhc, nsphc, ns;

  char junk_char[2];
  char temp_sequence[MAX_SEQ_LENGTH+2];

  CheckEssentials();

  nhc = num_hom_chromos;
  nsphc = num_strands_per_hom_chrom;
  ns = nhc * nsphc;

  for (extant_i = 0; extant_i < num_extant_total; extant_i++)
  {
    fscanf(fp_in, "%d", &ex_id);

    // also get the carriage return for the above line
    fgets(&(junk_char[0]), 2, fp_in);

    // first make the new node
    na_index = nt->node_id_ctr++;
    nt->NodeArray[na_index] = NewNode(na_index);

    nt->NodeArray[na_index]->external_node_id = ex_id;
    nt->NodeArray[na_index]->extant_leaf = 1;

    // see if this is the network outgroup and set up pointer if it is
    if (ex_id == nt->outgroup_external_id)
      nt->outgroup = nt->NodeArray[na_index];

    // now deal with all the sequences which need to each be on their own line
    hchromo_i = 0;
    strand_i = 0;

    for (seq_i = 0; seq_i < ns; seq_i++)
    {
      fgets(&(temp_sequence[0]), (sequence_length + 2), fp_in);
      strcpy(nt->NodeArray[na_index]->hom_chromos[hchromo_i]->strands[strand_i]->sequence, &(temp_sequence[0]));

      // replace the carriage return with a null character
      nt->NodeArray[na_index]->hom_chromos[hchromo_i]->strands[strand_i]->sequence[sequence_length] = '\0';

      // increment counters
      strand_i++;

      if (strand_i == num_strands_per_hom_chrom)
      {
	hchromo_i++;
        strand_i = 0;
      }
    }

    // up the counters
    nt->num_unique_nodes++;
    nt->num_unique_extant_leaves++;
  }

}
/*****************************************************************************/
/*                             Read Subtree Info                             */
/*        (the outgroup id, how many extant taxa impacted and which ones)    */
/*****************************************************************************/
void ReadSubtreeInfo(NETWORK nt)
{
  int found_yn;
  int id;
  int og, hi_num, pi_num, pi_num1, pi_num2;
  int num_lines, line_i;
  int i;
  int node_i;
  int subtree_processing_complete;
  int taxa_i;
  char str[MAX_LINE_LENGTH];
  char junk_char[2];

  // before proceeding make sure we have the extant taxa already
  // and that we are working on the first subtree -- remember only
  // one is permitted for this version!
  if (extant_taxa_read == 0)
  {
    printf("NR ERROR:  Cannot declare subtree info until extant taxa have been listed already -- program aborted\n");
    exit(0);
  }

  if (subtree_listing_ctr > 0)
  {
    printf("NR ERROR:  Limited to one hybrid subtree in this version -- program aborted\n");
    exit(0);
  }

  subtree_processing_complete = 0;

  // read outgroup info
  fscanf(fp_in, "%s", str);

  if (strcmp(str, "subtree_outgroup_id") == 0)
  {
    fscanf(fp_in, "%d", &og);
    nt->outgroup_ids_for_subtrees[subtree_listing_ctr] = og;
  }
  else
  {
    printf("NR ERROR:  Subtree listing format violated (must list subtree_outgroup_id not %s) -- program aborted\n", str);
    exit(0);
  }

  // read num of hybrid impacted extant taxa
  fscanf(fp_in, "%s", str);

  if (strcmp(str, "num_hybrid_impacted_taxa") == 0)
  {
    fscanf(fp_in, "%d", &hi_num);
  }
  else
  {
    printf("NR ERROR:  Subtree listing format violated (must list num_hybrid_impacted_taxa not %s) -- program aborted\n", str);
    exit(0);
  }

  // now handle those impacted taxa and the outgroup
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->external_node_id == 
        nt->outgroup_ids_for_subtrees[subtree_listing_ctr])
    {
      nt->outgroups_for_subtrees[subtree_listing_ctr] = nt->NodeArray[node_i];
    }
  }

  for (taxa_i = 0; taxa_i < hi_num; taxa_i++)
  {

    fscanf(fp_in, "%d", &id);

    if (id == og)
    {
      printf("NR ERROR: The outgroup (%d) for the subtree must be one of the non-impacted extant taxa -- program aborted\n", og);
      exit(0);
    }

    found_yn = 0;

    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      if (nt->NodeArray[node_i]->external_node_id == id)
      {
        // now we need to enter this subtree's/hybrid id into the first
	// non-used place in this node's impacted_by_hybrid array
	i = 0;

	while ( (nt->NodeArray[node_i]->impacted_by_hybrid[i] != -1) &&
	        (i < MAX_NUM_HYBRIDS) )
	{
	  i++;
	}

	// now i is where we want to put the value
	nt->NodeArray[node_i]->impacted_by_hybrid[i] = subtree_listing_ctr;

        // remember that we found the node
	found_yn = 1;
      } 
    }

    if (found_yn == 0)
    {
      printf("NR ERROR:  ID (%d) as hybrid impacted taxa for subtree %d not found -- program aborted\n", id, subtree_listing_ctr);
      exit(0);
    }
  }

  // get the carriage return
  fgets(&(junk_char[0]), 2, fp_in);

  // read method for identifying parental impacted taxa if present
  fscanf(fp_in, "%s", str);

  if (strcmp(str, "parental_offspring_id_method") == 0)
  {
    ReadParentalOffspringIdMethod();
  }
  else if (strcmp(str, "end_subtree_listing") == 0)
  {
    // we are done with subtree processing
    subtree_processing_complete = 1;
  }
  else
  {
    printf("NR ERROR:  Subtree listing not properly terminated -- program aborted\n");
    exit(0);
  }

  // deal with the 2 custom situation as they require more reading of input
  if ( (parental_offspring_id_method == 2) || 
       (parental_offspring_id_method == 3) )
  {
    fscanf(fp_in, "%s", str);

    if (strcmp(str, "num_parental_impacted_taxa") == 0)
    {
      fscanf(fp_in, "%d", &pi_num1);

      if (parental_offspring_id_method == 3)
      {
        fscanf(fp_in, "%d", &pi_num2);
      }
    }
    else
    {
      printf("NR ERROR:  Subtree listing format violated (must list num_parental_impacted_taxa not %s) -- program aborted\n", str);
    exit(0);
    }
  }

  // now read those taxa if dealing with one of the custom options
  if ( (parental_offspring_id_method == 2) ||
       (parental_offspring_id_method == 3) )
  {
    if (parental_offspring_id_method == 3)
    {
      num_lines = 2;
    }
    else
    {
      num_lines = 1;
    }

    for (line_i = 0; line_i < num_lines; line_i++)
    {
      if (line_i > 0) // get carriage return of previous, 
                      // but not the very last line of data as that
		      // is handled below
      {
        // get the carriage return
        fgets(&(junk_char[0]), 2, fp_in);
      }

      if (line_i == 0)
        pi_num = pi_num1;
      else
        pi_num = pi_num2;

      for (taxa_i = 0; taxa_i < pi_num; taxa_i++)
      {
        fscanf(fp_in, "%d", &id);

	// DEBUGGING
	//printf("The read in external taxon id is %d, method %d\n", id, parental_offspring_id_method);

        if (id == og)
        {
          printf("NR ERROR: The outgroup (%d) for the subtree must be one of the non-impacted (parental or hybrid) extant taxa\n", og);
          exit(0);
        }

        found_yn = 0;

        for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
        {
          if (nt->NodeArray[node_i]->external_node_id == id)
          {
            // now we need to enter this subtree's/hybrid id into the first
	    // non-used place in this node's impacted_by_parents array
	    i = 0;

	    while ( (nt->NodeArray[node_i]->impacted_by_parents[i] != -1) &&
	            (i < MAX_NUM_HYBRIDS) )
	    {
	      i++;
	    }

	    // now i is where we want to put the value
	    nt->NodeArray[node_i]->impacted_by_parents[i] = subtree_listing_ctr;

	    // also mark the parental affiliation in case we are doing the
	    // should give us -2 and -3 for values with the default at -1
	    // extreme custom option
	    if (parental_offspring_id_method == 3)
	    {
	      nt->NodeArray[node_i]->parental_aff = PARENTAL_AFF_DEFAULT - 
	                                            (line_i + 1);
	    }

            // remember that we found the node
	    found_yn = 1;
          } 
        }

        if (found_yn == 0)
        {
          printf("NR ERROR:  ID (%d) as parent impacted taxa for subtree %d not found -- program aborted\n", id, subtree_listing_ctr);
          exit(0);
        }
      }
    } // end of line reading for taxa
  } // end of custom(s) check

  // get the carriage return
  fgets(&(junk_char[0]), 2, fp_in);


  if (subtree_processing_complete == 0)
  {
    fscanf(fp_in, "%s", str);

    if (strcmp(str, "end_subtree_listing") == 0)
    {
      subtree_processing_complete = 1;
    }
    else
    {
      printf("NR ERROR:  Subtree listing was not properly terminated -- program aborted (not %s)\n", str);
      exit(0);
    }
  }

  // up the counter
  subtree_listing_ctr++;

  return;
}

/*****************************************************************************/
/*                      ReadParentalOffspringIdMethod                        */
/*****************************************************************************/
void ReadParentalOffspringIdMethod()
{
  char str[MAX_LINE_LENGTH];

  parental_offspring_id_method = 0;

  fscanf(fp_in, "%s", str);

  if ( (strcmp(str, "default") == 0) || (strcmp(str, "closest_neighbor") == 0))
  {
    parental_offspring_id_method = 0;
  }
  else if (strcmp(str, "closest_2x_hybrid") == 0)
  {
    parental_offspring_id_method = 1;
  }
  else if (strcmp(str, "custom") == 0)
  {
    parental_offspring_id_method = 2;
  }
  else if (strcmp(str, "extreme_custom") == 0)
  {
    parental_offspring_id_method = 3;
  }
  else
  {
    printf("NR ERROR:  Unknown method for identifying parental offspring (%s) -- program aborted\n", str);
    exit(0);
  }

  return;
}

/*****************************************************************************/
/*                         ReadDnaDistModel                                  */
/*****************************************************************************/
void ReadDnaDistModel()
{
  char str[MAX_LINE_LENGTH];
  char str2[DNA_DIST_CUSTOM_FN_STR_LENGTH-1];

  // read in the model information to be used for DNA dist routine
  // (dictates what dna dist opt file to make and/or use as the user
  // can specify and give a custom one...)
  
  fscanf(fp_in, "%s", str);

  if ( (strcmp(str, "jc") == 0) || (strcmp(str,"JC") == 0) )
  {
    dna_dist_model = 0;
  }
  else if ( (strcmp(str, "F84") == 0) || (strcmp(str,"f84") == 0) )
  {
    dna_dist_model = 1;
  }
  else if ( (strcmp(str, "k2p") == 0) || (strcmp(str,"K2P") == 0) )
  {
    dna_dist_model = 2;
  }
  else if (strcmp(str, "LogDet") == 0)
  {
    dna_dist_model = 3;
  }
  else if ( (strcmp(str, "custom") == 0) || (strcmp(str,"CUSTOM") == 0) )
  {
    dna_dist_model = 4;
    fgets(str2, (DNA_DIST_CUSTOM_FN_STR_LENGTH-1), fp_in);

    // get rid of newline character if present
    if (strchr(str2,'\n') != NULL)
    {
      strncpy(dna_dist_custom_file_name,str2,(strlen(str2)-1));
    }
    else
    {
      strcpy(dna_dist_custom_file_name,str2);
    }
  }
  else
  {
    printf("NR ERROR:  Invalid model name (%s) for DnaDist specified -- program aborted\n", str);
    exit(0);
  }

  return;
}



/*****************************************************************************/
/*                           Check Essentials                                */
/*****************************************************************************/
void CheckEssentials()
{
  int problem_yn;

  problem_yn = 0;

  if (sequence_length == 0)
  {
    printf("NR ERROR:  Sequence length must be declared BEFORE the extant taxa listing\n");
    problem_yn = 1;
  }
  else if (sequence_length > MAX_SEQ_LENGTH)
  {
    printf("NR ERROR:  Sequence length exceeds max sequence length of %d\n", MAX_SEQ_LENGTH);
    problem_yn = 1;
  }
  else if (num_extant_total == 0)
  {
    printf("NR ERROR:  Number of total extant taxa must be declared BEFORE the extant taxa listing\n");
    problem_yn = 1;
  }
  else if (num_hom_chromos == 0)
  {
    printf("NR ERROR:  Number of hom_chromos must be declared BEFORE the extant taxa listing\n");
    problem_yn = 1;
  }
  else if (num_strands_per_hom_chrom == 0)
  {
    printf("NR ERROR:  Number of strands per homologous chromosome must be declared BEFORE the extant taxa listing\n");
    problem_yn = 1;
  }
  else if ( (num_strands_per_hom_chrom % 2) != 0 )
  {
    printf("NR ERROR:  The number of strands per homologous chromosome MUST be even as we assume DIPLOID hybrids and there must be an even number to split between parents\n");
    problem_yn = 1;
  }
  else if ( (outgroup_adjustment != 0) && (outgroup_adjustment != 1) ) 
  {
    printf("NR ERROR:  Outgroup adjustment variable must be 0 or 1 (not %d).\n", outgroup_adjustment);
    problem_yn = 1;
  }
  else
  {
    problem_yn = 0;
  }
  
  if (problem_yn)
    exit(0);

  return;
}

/*****************************************************************************/
/*                        Check Post Essentials                              */
/*****************************************************************************/
void CheckPostEssentials(NETWORK nt)
{
  int problem_yn;

  problem_yn = 0;

  if ( (nt->outgroup == NULL) || 
       (nt->outgroup_ids_for_subtrees[0] == -1) )
  {
    printf("NR ERROR:  Both types (overall and subtree) outgroups must be declared\n");
    problem_yn = 1;
  }
  else if (nt->outgroups_for_subtrees[0] == NULL)
  {
    printf("NR ERROR:  Missing subtree outgroup\n");
    problem_yn = 1;
  }
  // make sure the outgroup is NOT an impacted node -- 
  // this test is redone when dealing with parental subtree
  else if ( (nt->outgroup->impacted_by_hybrid[0] != -1) ||
            (nt->outgroup->impacted_by_parents[0] != -1) ||
	    (nt->outgroups_for_subtrees[0]->impacted_by_hybrid[0] != -1) ||
	    (nt->outgroups_for_subtrees[0]->impacted_by_parents[0] != -1) )
  {
    printf("NR ERROR:  The outgroups cannot be impacted by a hybrid (or its parents)\n");
    problem_yn = 1;
  }

  if (subtree_listing_ctr >= MAX_NUM_HYBRIDS)
  {
    printf("NR ERROR:  Too many subtrees declared (%d) when the max allowed is %d\n", subtree_listing_ctr, MAX_NUM_HYBRIDS);

    problem_yn = 1;
  }
  else if (subtree_listing_ctr == 0)
  {
    printf("NR ERROR:  No subtree listing of impacted extant taxa -- no reason to use this code\n");

    problem_yn = 1;
  }
  else if (subtree_listing_ctr > 1)
  {
    printf("NR ERROR:  This delta version only handles ONE subtree\n");

    problem_yn = 1;
  }

  if ((recon_type == 1) && (mp_phylip_thresh != MP_PHYLIP_THRESH_DEFAULT))
  {
    printf("NR ERROR:  Maximum Parsimony threshold for PHYLIP is not applicable when RANDIP is the chosen reconstruction method -- program aborted\n");

    problem_yn = 1;
  }

  if (mp_phylip_thresh < 2)
  {
    printf("NR ERROR:  Maximum Parsimony threshold for PHYLIP must be at least 2 so that when the outgroup is added the minimum value of 3 is attained for PHYLIP's neighbor work -- program aborted\n");

    problem_yn = 1;
  }

  if (problem_yn)
    exit(0);

  return;
}
