/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<float.h>
#include<limits.h>
#include<math.h>
#include<time.h>
#include<ctype.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"twister.h"

/******************************************************************************/
/*                              NewNetwork                                    */
/******************************************************************************/

NETWORK NewNetwork() 
{
  int i;

  NETWORK nt = (NETWORK)malloc(sizeof(NETWORKTYPE));
 
  if (nt == NULL)
  {
    printf("NR ERROR: Could not allocate space for network\n");
    exit(0);
  }
  
  nt->root = NULL;
  nt->outgroup = NULL;
  nt->node_id_ctr = 0;
  nt->num_unique_nodes = 0;
  nt->num_unique_extant_leaves = 0;
  nt->num_reticulate_nodes = 0;
  nt->num_unique_reticulate_nodes = 0;
  nt->parsimony_score = -1;

  nt->outgroup_external_id = -1;

  // initialize outgroups and roots for subtrees
  for (i = 0; i < MAX_NUM_HYBRIDS; i++)
  {
    nt->outgroup_ids_for_subtrees[i] = -1;
    nt->subtree_roots[i] = NULL;
    nt->outgroups_for_subtrees[i] = NULL;
  }
  
  return nt;
}


/******************************************************************************/
/*                              CleanUpMemory                                 */
/******************************************************************************/
void CleanUpMemory(NETWORK nt)
{
  long node_i;

  // first clean up the network stuff
  for (node_i = 0; node_i < (nt->num_unique_nodes); node_i++)
  {
    free(nt->NodeArray[node_i]);
  }

  // now we are free to free the network -- and it has no dynamically
  // allocated structures left so just go ahead and free it!
  free(nt);

  return;
}

/******************************************************************************/
/*                                ParseNewick                                 */
/******************************************************************************/
void ParseNewick(NETWORK nt)
{
  int na_index;

  int properly_terminated;
  NODE current_node;
  NODE found_node;

  // make the root node and put it in the NodeArray
  na_index = nt->node_id_ctr++;
  nt->root = NewNode(na_index);
  nt->NodeArray[na_index] = nt->root;
  nt->num_unique_nodes++;

  // next we need to parse and create the appropriate network structure
  current_node = nt->root;

  while(GetToken() != EOF)
  {
    if (token_type == L_PAREN)
    {
      na_index = nt->node_id_ctr++;
      current_node->first_child_node = NewNode(na_index);
      nt->num_unique_nodes++;
      nt->NodeArray[na_index] = current_node->first_child_node;
      current_node->first_child_node->parent1_node = current_node;

      current_node = current_node->first_child_node;
    }
    else if (token_type == L_PAREN_TAXON_NAME)
    {
      //  in this situation we have an extant leaf which already
      //  exists in NodeArray so we must find it and create the
      //  appropriate relationships, but counters do not get 
      //  adjusted as we already have the node!
      
      found_node = FindNode(nt, token_int);
      current_node->first_child_node = found_node;
      found_node->parent1_node = current_node;

      current_node = found_node;
    }
    else if (token_type == COMMA_L_PAREN)
    {
      // we need to make the next sibling and hook it up correctly
      // we know this next node is an internal node
      na_index = nt->node_id_ctr++;
      current_node->next_sib_node = NewNode(na_index);
      current_node->next_sib_node->prev_sib_node = current_node;
      current_node->next_sib_node->parent1_node = current_node->parent1_node;

      nt->num_unique_nodes++;
      nt->NodeArray[na_index] = current_node->next_sib_node;

      // now move to the next sibling
      current_node = current_node->next_sib_node;
    }
    else if (token_type == COMMA_TAXON_NAME)
    {
      //  in this situation we have an extant leaf which already
      //  exists in NodeArray so we must find it and create the
      //  appropriate relationships, but counters do not get 
      //  adjusted as we already have the node!
      
      found_node = FindNode(nt, token_int);
      current_node->next_sib_node = found_node;
      found_node->prev_sib_node = current_node;
      found_node->parent1_node = current_node->parent1_node;

      // now move to that next sibling
      current_node = found_node;
    }
    else if (token_type == R_PAREN)
    {
      // move up to the parent
      current_node = current_node->parent1_node;
    }
    else if (token_type == R_PAREN_TAXON_NAME) // should not happen with PHYLIP
    {
      // move up to the parent and fill in the name
      current_node = current_node->parent1_node;
      current_node->external_node_id = token_int;
    }
    else if (token_type == POUND) // this case should not occur for PHYLIP
    {
      if (strcmp(token_char, "H") == 0)
      {
        current_node->reticulate = 1;
        nt->num_reticulate_nodes++;  // double counts, but we fix later
      }
      else if (strcmp(token_char, "LGT") == 0)
      {
        current_node->reticulate = 2;
        nt->num_reticulate_nodes++;  // double counts, but we fix later
      }
      else if (strcmp(token_char, "E") == 0)
      {
        current_node->extant_leaf = 1;
      }
      else
      {
        printf("NR ERROR: problem with reticulate node type, program aborted -- reticulate nodes should NOT have occured at this stage\n");
        exit(0);
      }
    }
    else if (token_type == COLON)
    {
      current_node->brlen1 = token_float;
    }
    else if (token_type == SEMI_COLON)
    {
      // we have the semicolon, but make sure we are at the root
      // which means having had the correct number of right parens
      if (current_node == nt->root)
      {
        properly_terminated = 1;
        break;  // we are done processing the topology
      }
      else
      {
        printf("NR ERROR: Reading the a tree was terminated prematurely, perhaps not enough )'s, program aborted\n");
	exit(0);
      }
    }
    else if (token_type == EOL)
    {
      printf("NR ERROR: we have an unexpected end of line while processing a consensus topology, program aborted\n");
      exit(0);
    }
    else if (token_type == ILLEGAL_CHAR)
    {
      printf("NR ERROR: illegal character (%d) encountered while parsing topology, program aborted\n", token_int);
      exit(0);
    }
    else if (token_type == UNKNOWN)
    {
      printf("NR ERROR: unknown character (%d) encountered while parsing consensus topology, program aborted\n", token_int);
      exit(0);
    }
    else
    {
      printf("NR ERROR: unidentified problem with parsing, program aborted\n");
      exit(0);
    }
  }
  
  return;
}

/******************************************************************************/
/*                                GetToken                                    */
/******************************************************************************/
int GetToken()
{
  int c;
  char *p = token_char;

  // NOTE: a carriage return can happen in many places, must check almost
  //       everywhere and allow parsing to continue!

  /* skip these */
  while (( c = getc(fp_temp_in)) == ' ' || c == '\t' || c  == '\n')
    ;

  if (c == '(')     /*************** LEFT PAREN ***************/
  {
    if ( (c = getc(fp_temp_in)) == '\n') ; else ungetc(c, fp_temp_in);
      
    // see if the next character is a taxon name or just another '('
    if ( (c = getc(fp_temp_in)) == '(')
    {
      ungetc(c, fp_temp_in);
      strcpy(token_char, "(");
      return token_type = L_PAREN;
    }
    else // we SHOULD have a taxon name -- which for our purposes must be an int
    {
      if (isdigit(c))
      {
	ungetc(c, fp_temp_in);
	fscanf(fp_temp_in, "%d", &c);
	token_int = c;

	return token_type = L_PAREN_TAXON_NAME;
      }
      else
      {
        token_int = c;
	return token_type = ILLEGAL_CHAR;
      }
    }
  }
  else if (c == ',')   /*************** COMMA **************/
  {
    if ( (c = getc(fp_temp_in)) == '\n') ; else ungetc(c, fp_temp_in);

    // see if the next character is another subtree, 
    // otherwise we should have a taxon name
    if ((c = getc(fp_temp_in)) == '(')
    {
      ungetc(c, fp_temp_in);
      strcpy(token_char, ",");
      return token_type = COMMA_L_PAREN;
    }
    else // we SHOULD have a taxon name -- which for our purposes must be an int
    {
      if (isdigit(c))
      {
        ungetc(c, fp_temp_in);
	fscanf(fp_temp_in, "%d", &c);
	token_int = c;

        return token_type = COMMA_TAXON_NAME;
      }
      else
      {
	token_int = c;
        return token_type = ILLEGAL_CHAR;
      }
	 
    }
  }
  else if (c == ')')            /**************  RIGHT PAREN  *************/
  {
    if ( (c = getc(fp_temp_in)) == '\n') ; else ungetc(c, fp_temp_in);

    // see if we have a taxon name following the right paren
    c = getc(fp_temp_in);
    if (isdigit(c))
    {
      // read the sub-root's name -- which for our purposes must be an int
      ungetc(c, fp_temp_in);
      fscanf(fp_temp_in, "%d", &c);
      token_int = c;

      return token_type = R_PAREN_TAXON_NAME;
    } 
    else
    {
      ungetc(c, fp_temp_in);
      strcpy(token_char, ")");
      return token_type = R_PAREN;
    }
  }
  /*this should not be used by consense as PHYLIP does not know about networks*/
  else if (c == '#')            /**************  POUND  ********************/
  {
    if ( (c = getc(fp_temp_in)) == '\n') ; else ungetc(c, fp_temp_in);

    // next letters must specify type of reticulate node 
    // either H for hybrid or LGT for lateral gene transfer
    // OR extant taxa with an E
    c = getc(fp_temp_in);

    if ( (c == 'H') || (c == 'h') )
    {
      strcpy(token_char, "H");
      return token_type = POUND;
    }
    else if ( (c == 'E') || (c = 'e') )
    {
      strcpy(token_char, "E");
      return token_type = POUND;
    }
    else if ( (c == 'L' || c == 'l') &&
	      ((c = getc(fp_temp_in)) == 'G' || c == 'g') &&
	      ((c = getc(fp_temp_in)) == 'T' || c == 't') )
    {
      strcpy(token_char, "LGT");
      return token_type = POUND;
    }
    else
    {
      token_int = c;
      return token_type = UNKNOWN;
    }
  }
  else if (c == ':')            /**************  COLON  ********************/
  {
    if ( (c = getc(fp_temp_in)) == '\n') ; else ungetc(c, fp_temp_in);

    // PHYLIP does give negative branch lengths -- watch out!!!
    if ( ! ( (isdigit(c = getc(fp_temp_in))) || (c == '-') )  )
    {
      token_int = c;
      return token_type = ILLEGAL_CHAR;
    }
    else
    {
      // read the branch length info
      for (*p++ = c; isdigit(c= getc(fp_temp_in))||(c == '.')||(c == '-'); )
        *p++ = c;
      *p = '\0';
      // if we had a branch length for the root -- problem!
      if (c == ';')
      {
	token_int = c;
        return token_type = ILLEGAL_CHAR;
      }
      else
      {
        ungetc(c, fp_temp_in);
        token_float = atof(token_char);
        return token_type = COLON;
      }
    }
  }
  else if (c == ';')            /**************  SEMI COLON  ***************/
  {
    strcpy(token_char, ";");
    return token_type = SEMI_COLON;
  }
  else
  {
    token_int = c;
    return token_type = UNKNOWN;/**************  UNKNOWN *******************/
  }
}

/******************************************************************************/
/*                           DownwardAssignRound                              */
/******************************************************************************/
void DownwardAssignRound(NODE nd, int base_index, int hchromo_i, int strand_i)
{
  int rand_try_ctr, rand_problem, rand_val;
  int chosen_base, majority_base;
  int num_node_kids;
  NODE next_node;

  // NOTE:  the fixed sequence check makes sure the sequence has not
  //        been assigned as part of the reticulate work and hybrid
  //        subtree work as those seqs should not be overwritten

  /***********************/
  /***********************/
  /**     root case     **/
  /***********************/
  /***********************/
  if ( (nd->parent1_node == NULL) && (nd->parent2_node == NULL) )
  {
    if (nd->hom_chromos[hchromo_i]->strands[strand_i]->fixed_seq == 0)
    {
      // first see if the root has more than two kids -- if so we
      // need to do the assignment differently 
      num_node_kids = CountKids(nd);

      if (num_node_kids > 2)
      {
        majority_base = ReturnMajorityBase(nd, NULL, base_index, hchromo_i, strand_i);
        // reset all values
        ResetAllBaseValues(nd, base_index, hchromo_i, strand_i);
        // set the majority value
        SetSingleBaseValue(nd, base_index, hchromo_i, strand_i, majority_base);
        // set the corresponding character
        SetBaseChar(nd, base_index, hchromo_i, strand_i, majority_base);

	//printf("NR WARNING:  We had the multiple kid situation for base assignment occur (with the root node) -- majority base was %d\n", majority_base);
      }
      else // if two kids or less do the regular random assingment stuff
      {
        rand_try_ctr = 0;
        rand_problem = 1;

        while ( (rand_problem == 1) && (rand_try_ctr < 1000) )
        {
          // this does a rand value in the range of [0,1) then multplied by
          // four for the number of bases which randomly selects a base
          // of 0-3 which is what we need here
          rand_val = (int) (genrand_real2() * 4);

          if (nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[rand_val][base_index] == 1)
          {
            // we have our root base, so...
            // reset all values
	    ResetAllBaseValues(nd, base_index, hchromo_i, strand_i);
	    // set the chosen value
	    SetSingleBaseValue(nd, base_index, hchromo_i, strand_i, rand_val);
	    // set the corresponding character
	    SetBaseChar(nd, base_index, hchromo_i, strand_i, rand_val);
	    // we are done!
            rand_problem = 0;
	    break;
          }
        }

        // safety catch if something has gone horribly wrong above
        if (rand_problem == 1)
        {
          printf("NR ERROR: Could not select a base for the root at index %d\n", base_index);
          exit(0);
        }
      } // end of kid count check for root
    } // end of seq work if seq was NOT fixed

    // now recurse for all nodes below the root -- note that
    // immediate (1 step) descendants from the root cannot be a
    // reticualte node so do not need to worry about that here
    if (nd->first_child_node != NULL)
    {
      next_node = nd->first_child_node;

      while(next_node != NULL)
      {
        DownwardAssignRound(next_node, base_index, hchromo_i, strand_i);
	next_node = next_node->next_sib_node;
      }
    }
  } // end of root case

  /***********************/
  /***********************/
  /**  reticulate case  **/
  /***********************/
  /***********************/
  else if (nd->reticulate == 1) 
  {
    ; // do nothing
  }

  /***********************/ 
  /***********************/
  /**    regular case   **/
  /***********************/
  /***********************/
  else if ( (nd->parent1_node != NULL) && 
            (nd->parent2_node == NULL) &&
	    (nd->extant_leaf == 0) ) // don't bother with extant leaves
  {
    if (nd->hom_chromos[hchromo_i]->strands[strand_i]->fixed_seq == 0)
    {
      // first see if this node has more than two kids -- if so we
      // need to do the assignment differently 
      num_node_kids = CountKids(nd);

      if (num_node_kids > 2)
      {
        majority_base = ReturnMajorityBase(nd, nd->parent1_node, base_index, hchromo_i, strand_i);

        // reset all values
        ResetAllBaseValues(nd, base_index, hchromo_i, strand_i);
        // set the majority value
        SetSingleBaseValue(nd, base_index, hchromo_i, strand_i, majority_base);
        // set the corresponding character
        SetBaseChar(nd, base_index, hchromo_i, strand_i, majority_base);

	//printf("NR WARNING:  We had the multiple kid situation for base assignment occur (int node id %d) -- majority base was %d\n", nd->internal_node_id, majority_base);
      }
      else
      {
        // note: chosen base returns a -1 if the value is a ?

        chosen_base = ReturnDesiredBase(nd->parent1_node, base_index, hchromo_i, strand_i);
      
        // now see if the current node has the same base as an option
        // if yes, get it set up for this node as well
        if ( (chosen_base != -1) &&
             (nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[chosen_base][base_index] == 1) )
        {
          // reset all values
          ResetAllBaseValues(nd, base_index, hchromo_i, strand_i);
          // set the chosen value
          SetSingleBaseValue(nd, base_index, hchromo_i, strand_i, chosen_base);
          // set the corresponding character
          SetBaseChar(nd, base_index, hchromo_i, strand_i, chosen_base);
        }
        else // parent's chosen base is not an option for this node
             // (or parent has a ? meaning any value is fine)
             // so randomly choose a base that works for this node
        {
          rand_try_ctr = 0;
          rand_problem = 1;

          while ( (rand_problem == 1) && (rand_try_ctr < 1000) )
          {
            // this does a rand value in the range of [0,1) then multplied by
            // four for the number of bases which randomly selects a base
            // of 0-3 which is what we need here
            rand_val = (int) (genrand_real2()*4);

            if (nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[rand_val][base_index] == 1)
            {
              // reset all values
              ResetAllBaseValues(nd, base_index, hchromo_i, strand_i);
              // set the chosen value
              SetSingleBaseValue(nd, base_index, hchromo_i, strand_i, rand_val);
              // set the corresponding character
	      SetBaseChar(nd, base_index, hchromo_i, strand_i, rand_val);
	      // we are done!
              rand_problem = 0;
	      break;
            }
          }

          // safety catch if something has gone horribly wrong above
          if (rand_problem == 1)
          {
            printf("NR ERROR: Could not select a base for the root at index %d\n", base_index);
            exit(0);
          }
        } // parent's choice was not an option
      } // kid count check
    } // end of assignment work if seq is NOT fixed already

    // now recurse for all nodes below this node
    if (nd->first_child_node != NULL)
    {
      if (nd->first_child_node->reticulate == 1)
      {
        if (nd->first_child_node->parent1_node == nd)
	{
          next_node = nd->first_child_node->next_p1_sib;
	}
	else if (nd->first_child_node->parent2_node == nd)
	{
          next_node = nd->first_child_node->next_p2_sib;
	}
	else
	{
	  printf("NR ERROR: Fatal problem processing reticulate node during downward assignment -- program aborted\n");
	  exit(0);
	}

        while(next_node != NULL)
        {
          DownwardAssignRound(next_node, base_index, hchromo_i, strand_i);
	  next_node = next_node->next_sib_node;
        }
      }
      else // regular first child node
      {
        next_node = nd->first_child_node;

        while(next_node != NULL)
        {
          DownwardAssignRound(next_node, base_index, hchromo_i, strand_i);
	  next_node = next_node->next_sib_node;
        }
      }
    }
  } // regular case
  else
  {
    // should have been processed above
    if ( (nd->extant_leaf == 0) && (nd->reticulate == 0) &&
         (nd->hom_chromos[hchromo_i]->strands[strand_i]->fixed_seq == 0) )
    {
      printf("NR ERROR: Something is up with the downward assignment, program aborted");
      exit(0);
    }
  }

  return;
}

/******************************************************************************/
/*                            UpwardOptionRound                               */
/******************************************************************************/
void UpwardOptionRound(NODE nd, int base_index, int hchromo_i, int strand_i)
{
  NODE next_node;

  int empty_intersection;

  // Do the post-order traversal work here. In general, we do not
  // want to deal with reticulate nodes so check for that.  If
  // this routine is called for the hybrid subtree root, that
  // node is not set to reticulate until later so this still works.
  // Also note that we need to proceed with the hybrid's siblings
  // (either via parent1 or parent2) as those offspring need seqs.
  // Also, if a seq is fixed, it does not get options, but we
  // still must proceed to its offspring. 
  
  if ( (nd->reticulate == 0) && (nd->first_child_node != NULL) ) 
  {
    next_node = nd->first_child_node;

    if (next_node->reticulate == 1)
    {
      if (next_node->parent1_node == nd)
      {
        last_hyb_parent_hit = 1;
      }
      else if (next_node->parent2_node == nd)
      {
        last_hyb_parent_hit = 2;
      }
      else
      {
        printf("NR ERROR: Problem with reticulate node in UpwardOptionRound (hyb_parent_hit_ctr %d  and last_hyb_parent_hit %d) -- program aborted\n", hyb_parent_hit_ctr, last_hyb_parent_hit);
	exit(0);
      }
    }

    while(next_node != NULL)
    {
      UpwardOptionRound(next_node, base_index, hchromo_i, strand_i);
      next_node = next_node->next_sib_node;
    }
  }
  else if ( (nd->reticulate == 1) && (last_hyb_parent_hit == 1) )
  {
    hyb_parent_hit_ctr++;

    if (nd->next_p1_sib != NULL)
    {
      next_node = nd->next_p1_sib;

      while(next_node != NULL)
      {
        UpwardOptionRound(next_node, base_index, hchromo_i, strand_i);
        next_node = next_node->next_sib_node;
      }
    }
  }
  else if ( (nd->reticulate == 1) && (last_hyb_parent_hit == 2) )
  {
    hyb_parent_hit_ctr++;

    if (nd->next_p2_sib != NULL)
    {
      next_node = nd->next_p2_sib;

      while(next_node != NULL)
      {
        UpwardOptionRound(next_node, base_index, hchromo_i, strand_i);
        next_node = next_node->next_sib_node;
      }
    }
  }
  else if (hyb_parent_hit_ctr > 2)
  {
    printf("NR ERROR: problem with traversal in upward option round (internal node id %d  parental hit ctr is %d) -- program aborted\n", nd->internal_node_id, hyb_parent_hit_ctr);
    exit(0);
  }
  else if ( (nd->reticulate == 0) && (nd->first_child_node == NULL) ) 
  {
   ; // do nothing
  }
  else
  {
    ; // do nothing, it is possible the reticulate node does not have any sibs
  }


  // do what we need to do here for the "merging" as all the necessary
  // traversal for this node (namely its descendants) has been done
  // already through the above recursive calls
   
  // current node has descendants and therefore needs assigning provided
  // it is NOT a reticulate node -- descendants of reticulate node have
  // their seqs as marked, so should skip doing those as well
  if (nd->reticulate == 0)
  {
    if ( (nd->hom_chromos[hchromo_i]->strands[strand_i]->fixed_seq == 0) &&
         (nd->first_child_node != NULL) )
    {
      // if this node only has ONE immediate descendant, set this 
      // node's base options to that of its descendant's automatically
      // (note: must be careful to establish really only ONE descendant!)
      if (  ((nd->first_child_node->reticulate == 0) &&
             (nd->first_child_node->next_sib_node == NULL)) ||
	    ((nd->first_child_node->reticulate == 1) &&
	     (nd->first_child_node->parent1_node == nd) &&
	     (nd->first_child_node->next_p1_sib == NULL)) ||
	    ((nd->first_child_node->reticulate == 1) &&
	     (nd->first_child_node->parent2_node == nd) &&
	     (nd->first_child_node->next_p2_sib == NULL)) )
      {
        SetAllPosBaseValues(nd, nd->first_child_node, base_index, hchromo_i, strand_i);
      }
      else // more than 1 desc
      {
        // if this node has more than one descendant -- first we need to see
        // if the intersection of all descendant bases is empty or not
        empty_intersection = CheckForEmptyDescInter(nd, base_index, hchromo_i, strand_i);

        if (empty_intersection)
        {
          DoDescUnion(nd, base_index, hchromo_i, strand_i);
        }
        else 
        {
          DoDescIntersection(nd, base_index, hchromo_i, strand_i);
        }
      }
    }
  }

  return;
}

/******************************************************************************/
/*                         PopulateExtantBaseValues                           */
/******************************************************************************/
void PopulateExtantBaseValues(NODE nd)
{
  // this is where we populate the options array for the given node
  // also, as a safety catch we mark the sequence as fixed so it will
  // not get over-written when processed in downward work -- important
  // for parental hybrid nodes

  int hchromo_i, strand_i, base_i, fixed;
  char b[2]; // because strcpy insists on adding a '\0'

  // assume fixed unless we hit a ? mark -- our seqs at the strand
  // level are either all assigned or all ?
  fixed = 1;

  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      // assume fixed unless we hit a ? mark -- our seqs at the 
      // strand level are either all assigned or all ?
      fixed = 1;

      for (base_i = 0; base_i < sequence_length; base_i++)
      {
        // set what the base is
        if (strncmp(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_i]), "A", 1) == 0)
	  strcpy(&(b[0]), "A");
        else if (strncmp(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_i]), "G", 1) == 0)
	  strcpy(&(b[0]), "G");
        else if (strncmp(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_i]), "C", 1) == 0)
	  strcpy(&(b[0]), "C");
        else if (strncmp(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_i]), "T", 1) == 0)
	  strcpy(&(b[0]), "T");
        else if (strncmp(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_i]), "?", 1) == 0)
	  strcpy(&(b[0]), "?");
	else
	{
	  printf("NR ERROR: unexpected error in populate extant base values routine -- program aborted\n");
	  exit(0);
	}
 
        if (strcmp(&(b[0]), "A") == 0)
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[0][base_i] = 1;
        else if (strcmp(&(b[0]), "G") == 0)
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[1][base_i] = 1;
        else if (strcmp(&(b[0]), "C") == 0) 
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[2][base_i] = 1;
        else if (strcmp(&(b[0]), "T") == 0) 
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[3][base_i] = 1;
        else if (strcmp(&(b[0]), "?") == 0) // any base option 
	{
	  fixed = 0;

          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[0][base_i] = 1;
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[1][base_i] = 1;
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[2][base_i] = 1;
          nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[3][base_i] = 1;
	}
	else
	{
	  printf("NR ERROR: unexpected error in populate extant base values routine -- program aborted\n");
	  exit(0);
	}

      } // next base index (base_i)

      if (fixed == 1)
        nd->hom_chromos[hchromo_i]->strands[strand_i]->fixed_seq = 1;
      else
        nd->hom_chromos[hchromo_i]->strands[strand_i]->fixed_seq = 0;

    } // next strand_i
  } // next hchromo_i

  return;

}

/******************************************************************************/
/*                         SetAllPosBaseValues                                */
/******************************************************************************/
void SetAllPosBaseValues(NODE p_node, NODE c_node, int base_index, int hchromo_i,
                         int strand_i)
{
  // this is where we "turn on" (1) all the options for the parent based on
  // what the child considers possible -- note, we only do the positives
  // here (cannot reset any of the parent options to 0 as it may be possible
  // for the parent based on other children)

  int c_node_a;
  int c_node_g;
  int c_node_c;
  int c_node_t;

  /********************************/
  /*   check the child's values   */
  /********************************/
  c_node_a = c_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[0][base_index];

  c_node_g = c_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[1][base_index];

  c_node_c = c_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[2][base_index];

  c_node_t = c_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[3][base_index];

  /***************************************************************************/
  /* now if any of the above are 1, set the parent's same value to 1 as well */
  /***************************************************************************/
  if (c_node_a == 1)
    p_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[0][base_index] = 1;

  if (c_node_g == 1)
    p_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[1][base_index] = 1;

  if (c_node_c == 1)
    p_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[2][base_index] = 1;

  if (c_node_t == 1)
    p_node->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[3][base_index] = 1;

  return;
}

/******************************************************************************/
/*                         CheckForEmptyDescInter                             */
/******************************************************************************/
int CheckForEmptyDescInter(NODE nd, int base_index, int hchromo_i, int strand_i)
{
  int base_i, is_empty;
  int desc_ctr;
  int parent;
  int base_ctr_array[4];
  NODE tmp_nd;

  // in this routine we determine if the intersection of the descendant
  // base option arrays is empty or not....

  // intialize our variables
  desc_ctr = 0;
  is_empty = 1;
  parent = -1;
  tmp_nd = nd->first_child_node;

  for (base_i = 0; base_i < 4 ; base_i++)
  {
    base_ctr_array[base_i] = 0;
  }

  // if we are dealing with one of two parents to a hyrid node, we need
  // to record which parent we are dealing with here....
  if (nd->first_child_node->reticulate == 1)
  {
    if (nd->first_child_node->parent1_node == nd)
      parent = 1;

    if (nd->first_child_node->parent2_node == nd)
      parent = 2;
  }

  // now do the counting
  while (tmp_nd != NULL)
  {
    for (base_i = 0; base_i < 4; base_i++)
    {
      if (tmp_nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_i][base_index] == 1)   
      {
        base_ctr_array[base_i] = base_ctr_array[base_i] + 1;
      }
    }
    
    desc_ctr++;

    // now get the next node but be careful as the first child may be 
    // a reticulate node and the next sib is stored differently in that case
    if (tmp_nd->reticulate == 1)
    {
      if (parent == 1)
      {
        tmp_nd = tmp_nd->next_p1_sib;
      }
      else if (parent == 2) 
      {
        tmp_nd = tmp_nd->next_p2_sib;
      }
      else
      {
        printf("NR ERROR: Unexpected issue with reticulate node in CheckForEmptyDescInter routine -- program aborted\n");
	exit(0);
      }
    }
    else
    {
      tmp_nd = tmp_nd->next_sib_node;
    }
  }

  if (desc_ctr < 2)
  {
    printf("NR ERROR: CheckForEmptyDescInter routine was called on a node (internal id %d and %d descendant(s)) with less than two descendants, this should not have occurred -- program aborted\n", nd->internal_node_id, desc_ctr);
    exit(0);
  }

  // now see if the intersection is empty or not....for it to be non-empty
  // at least one of the values in the array must be the same as the
  // desc_ctr
  for (base_i = 0; base_i < 4 ; base_i++)
  {
    if (base_ctr_array[base_i] == desc_ctr)
      is_empty = 0;
  }

  return is_empty;
}

/******************************************************************************/
/*                                DoDescUnion                                 */
/******************************************************************************/
void DoDescUnion(NODE nd, int base_index, int hchromo_i, int strand_i)
{
  NODE tmp_nd;
  int base_i;

  // this is where the given node gets all the possible bases from all
  // of its descendants

  tmp_nd = nd->first_child_node;

  while(tmp_nd != NULL)
  {
    for (base_i = 0; base_i < 4; base_i++)
    {
      if (tmp_nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_i][base_index] == 1) 
      {
        nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_i][base_index] = 1;
      }

    }

    // look at the next desc for tmp_nd but be careful if this is a reticulate
    // node, its first sibling is stored in a different way.....
    if (tmp_nd->reticulate == 1)
    {
      if (tmp_nd == tmp_nd->parent1_node->first_child_node)
      {
        tmp_nd = tmp_nd->next_p1_sib;
      }
      else if (tmp_nd == tmp_nd->parent2_node->first_child_node)
      {
        tmp_nd = tmp_nd->next_p2_sib;
      }
      else
      {
        printf("NR ERROR: Unexpected error with reticulate node in DoDescUnion -- program aborted\n");
	exit(0);
      }
    }
    else
    {
      tmp_nd = tmp_nd ->next_sib_node;
    }
  }

  return;
}

/******************************************************************************/
/*                             DoDescIntersection                             */
/******************************************************************************/
void DoDescIntersection(NODE nd, int base_index, int hchromo_i, int strand_i)
{
  NODE tmp_nd;
  int base_i;
  int desc_ctr;
  int base_ctr_array[4];

  // this is where the given node gets the intersection of bases from
  // all of its descendants -- initial code is like the EmptyIntersection
  // check, but the outcome is handled differently
  
  // intialize our variables
  desc_ctr = 0;
  tmp_nd = nd->first_child_node;

  for (base_i = 0; base_i < 4 ; base_i++)
  {
    base_ctr_array[base_i] = 0;
  }

  // now do the counting
  while (tmp_nd != NULL)
  {
    for (base_i = 0; base_i < 4; base_i++)
    {
      if (tmp_nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_i][base_index] == 1)   
      {
        base_ctr_array[base_i] = base_ctr_array[base_i] + 1;
      }
    }
    
    desc_ctr++;

    // look at the next desc for tmp_nd but be careful if this is a reticulate
    // node, its first sibling is stored in a different way.....
    if (tmp_nd->reticulate == 1)
    {
      if (tmp_nd == tmp_nd->parent1_node->first_child_node)
      {
        tmp_nd = tmp_nd->next_p1_sib;
      }
      else if (tmp_nd == tmp_nd->parent2_node->first_child_node)
      {
        tmp_nd = tmp_nd->next_p2_sib;
      }
      else
      {
        printf("NR ERROR: Unexpected error with reticulate node in DoDescUnion -- program aborted\n");
	exit(0);
      }
    }
    else
    {
      tmp_nd = tmp_nd->next_sib_node;
    }
  }

  // now see for which ones there is an intersection and set the base
  // for the given node accordingly
  for (base_i = 0; base_i < 4; base_i++)
  {
    if (base_ctr_array[base_i] == desc_ctr)
    {
      nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[base_i][base_index] = 1;
    }
  }

  return;
}

/******************************************************************************/
/*                         ResetAllBaseValues                                 */
/******************************************************************************/
void ResetAllBaseValues(NODE nd, int base_index, int hchromo_i, int strand_i)
{
  nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[0][base_index] = 0;
  nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[1][base_index] = 0;
  nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[2][base_index] = 0;
  nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[3][base_index] = 0;

  return;
}

/******************************************************************************/
/*                           SetSingleBaseValue                               */
/******************************************************************************/
void SetSingleBaseValue(NODE nd, int base_index, int hchromo_i, int strand_i, 
                       int chosen_base)
{
  nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[chosen_base][base_index] = 1;

  return;
}

/******************************************************************************/
/*                             SetBaseChar                                    */
/******************************************************************************/
void SetBaseChar(NODE nd, int base_index, int hchromo_i, int strand_i, 
                 int chosen_base)
{
  if (chosen_base == 0)
    strncpy(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "A", 1);

  if (chosen_base == 1)
    strncpy(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "G", 1);

  if (chosen_base == 2)
    strncpy(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "C", 1);

  if (chosen_base == 3)
    strncpy(&(nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "T", 1);

  return;
}

/******************************************************************************/
/*                              ReturnDesiredBase                             */
/******************************************************************************/
int ReturnDesiredBase(NODE parent_nd, int base_index, int hchromo_i,int strand_i)
{
  int desired_base;

  desired_base = -5;

  if (strncmp(&(parent_nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "?", 1) == 0)
    desired_base = -1;

  if (strncmp(&(parent_nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "A", 1) == 0)
    desired_base = 0;

  if (strncmp(&(parent_nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "G", 1) == 0)
    desired_base = 1;

  if (strncmp(&(parent_nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "C", 1) == 0)
    desired_base = 2;

  if (strncmp(&(parent_nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence[base_index]), "T", 1) == 0)
    desired_base = 3;

  if (desired_base == -5)
  {
    printf("NR ERROR: Something is amiss with downward assignment in return desired base routine (internal id for parent is %d -- bi %d hchromo %d strand %d seq '%s')\n", parent_nd->internal_node_id, base_index, hchromo_i, strand_i, parent_nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence);
    exit(0);
  }

  return desired_base;
}

/******************************************************************************/
/*                             CountKids                                      */
/******************************************************************************/
int CountKids(NODE nd)
{
  int count;
  NODE tmp_nd;

  // original nd must NOT be reticulate when calling this routine
  if (nd->reticulate != 0)
  {
    printf("NR ERROR:  Count kids was called for a reticulate node -- not good -- program aborted\n");
    exit(0);
  }

  count = 0;

  // only the first child node may be reticulate....
  if (nd->first_child_node != NULL)
  {
    count++;

    if (nd->first_child_node->reticulate != 0)
    {
      if (nd->first_child_node->parent1_node == nd)
        tmp_nd = nd->first_child_node->next_p1_sib;
      else
        tmp_nd = nd->first_child_node->next_p2_sib;
    }
    else
    {
      tmp_nd = nd->first_child_node->next_sib_node;
    }

    while (tmp_nd != NULL)
    {
      count++;
      tmp_nd = tmp_nd->next_sib_node;
    }
  }

  return count;
}

/******************************************************************************/
/*                             ReturnMajorityBase                             */
/******************************************************************************/
int ReturnMajorityBase(NODE nd, NODE parent, int base_index, int hchromo_i, 
                       int strand_i)
{
  int maj_base, chosen;
  int base_ctr_array[4];
  int i, max_val, ctr, max_tries;
  NODE tmp_nd;

  // this routine examines the offspring and parent (if not null) of the
  // given node and sees which base is most popular -- if there is a tie
  // a random selection of those tied is performed....

  // inialize counter array
  for (i = 0; i < 4; i++)
  {
     base_ctr_array[i] = 0; 
  }

  // count stuff for the descendants of the node
  tmp_nd = nd->first_child_node;

  while (tmp_nd != NULL)
  {
    for (i = 0; i < 4; i++)
    {
      if (tmp_nd->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[i][base_index] == 1)
      {
        base_ctr_array[i] = base_ctr_array[i] + 1;
      }
    }

    if (tmp_nd->reticulate != 0)
    {
      if (tmp_nd->parent1_node == nd)
        tmp_nd = tmp_nd->next_p1_sib;
      else if (tmp_nd->parent2_node == nd)
        tmp_nd = tmp_nd->next_p2_sib;
      else
      {
        printf("NR ERROR:  unexpected problem in return majority base -- program aborted\n");
	exit(0);
      }
    }
    else
    {
      tmp_nd = tmp_nd->next_sib_node;
    }
  }

  // if parent is not null, add in the counts for it too...
  if (parent != NULL)
  {
    for (i = 0; i < 4; i++)
    {
      if (parent->hom_chromos[hchromo_i]->strands[strand_i]->base_option_array[i][base_index] == 1)
      {
        base_ctr_array[i] = base_ctr_array[i] + 1;
      }
    }
  }
   
  max_val = MaxInt(base_ctr_array[0], base_ctr_array[1], base_ctr_array[2], base_ctr_array[3]);
  ctr = 0;
  maj_base = -1;

  // see if more than one index has the max value
  for (i = 0; i < 4; i++)
  {
    if (base_ctr_array[i] == max_val)
    {
      ctr++;
      maj_base = i;
    }
  }

  if (ctr == 0)
  {
    printf("NR ERROR: problem in return major base (ctr was 0) -- program aborted\n");
    exit(0);
  }
  else if (ctr == 1)
  {
    return maj_base; // was set above
  }
  else if (ctr > 1) // tie
  {
    maj_base = -1;
    max_tries = 0;
    
    while ( (maj_base == -1) && (max_tries < 1000) )
    {
      chosen = (int) (genrand_real2() * 4);

      if (base_ctr_array[chosen] == max_val)
        maj_base = chosen;
      
      max_tries++;
    }

    if (maj_base == -1)
    {
      printf("NR ERROR: problem in return major base with tie logic -- program aborted\n");
      exit(0);
    }

  }
  else 
  {
    printf("NR ERROR: problem in return major base -- program aborted\n");
    printf("DEBUG INFO:  ctr is %d and the bcarray values are %d %d %d %d with max having been %d\n", ctr, base_ctr_array[0], base_ctr_array[1], base_ctr_array[2], base_ctr_array[3], max_val);
    exit(0);
  }

  return maj_base;
}

/******************************************************************************/
/*                          CountExtantImpactedByHybrid                       */
/******************************************************************************/
int CountExtantImpactedByHybrid(NETWORK nt)
{
  int node_i;
  int ctr;

  // WARNING:  remember only one hybrid and its parents!

  ctr = 0;

  // go through node array and up counter if node is impacted 
  // by the hybrid 

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if ( (nt->NodeArray[node_i]->impacted_by_hybrid[0] != -1) &&
         (nt->NodeArray[node_i]->extant_leaf == 1) )
    {
      ctr++;
    }
  }

  return ctr;
}

/******************************************************************************/
/*                          CountExtantImpactedByParents                      */
/******************************************************************************/
int CountExtantImpactedByParents(NETWORK nt)
{
  int node_i;
  int ctr;

  // WARNING:  remember only one hybrid and its parents!

  ctr = 0;

  // go through node array and up counter if node is impacted 
  // by the parents of the hybrid

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if ( (nt->NodeArray[node_i]->impacted_by_parents[0] != -1) &&
         (nt->NodeArray[node_i]->extant_leaf == 1) )
    {
      ctr++;
    }
  }

  return ctr;
}

/******************************************************************************/
/*                          CountExtantImpactedByParent                       */
/******************************************************************************/
int CountExtantImpactedByParent(NETWORK nt, int paffval)
{
  int node_i;
  int ctr;

  // WARNING:  remember only one hybrid and A single parent!

  ctr = 0;

  // go through node array and up counter if node is impacted/affiliated
  // with the provided parent 

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if ( (nt->NodeArray[node_i]->parental_aff == paffval) &&
         (nt->NodeArray[node_i]->extant_leaf == 1) )
    {
      ctr++;
    }
  }

  return ctr;
}


/******************************************************************************/
/*                           CountExtantNotImpacted                           */
/******************************************************************************/
int CountExtantNotImpacted(NETWORK nt)
{
  int node_i;
  int ctr;

  // WARNING:  remember only one hybrid and its parents!

  ctr = 0;

  // go through node array and up counter if node is NOT impacted 
  // by the hybrid or its parents

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if ( (nt->NodeArray[node_i]->impacted_by_hybrid[0] == -1) &&
         (nt->NodeArray[node_i]->impacted_by_parents[0] == -1) &&
         (nt->NodeArray[node_i]->extant_leaf == 1) )
    {
      ctr++;
    }
  }

  return ctr;
}

/******************************************************************************/
/*                               CountImpacted                                */
/******************************************************************************/
int CountImpacted(NETWORK nt)
{
  int node_i;
  int ctr;

  ctr = 0;

  // go through node array and up counter if node is impacted by at least
  // one hybrid

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->impacted_by_hybrid[0] != -1)
    {
      ctr++;
    }
  }

  return ctr;

}

/******************************************************************************/
/*                        Tree Hunting Related Functions                      */
/******************************************************************************/
int IsTree(NETWORK nt)
{
  int node_i;
  int result;

  result = 1;

  // go through all the nodes and see if we have any reticulate nodes
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->reticulate != 0)
      result = 0;
  }

  return result;
}

/******************************************************************************/
/*                          ResetMarkersAll                                   */
/******************************************************************************/
void ResetMarkersAll(NETWORK nt)
{
  int node_i;

  // This routine resets ALL node markers by going through the node array

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    nt->NodeArray[node_i]->marker = 0;
  }

  return;

}

/******************************************************************************/
/*                           Reset Markers Topo                               */
/******************************************************************************/
void ResetMarkersTopo(NODE nd)
{
  hyb_parent_hit_ctr = 0;
  last_hyb_parent_hit = 0;

  ResetMarkersTopoHelper(nd);
}

/******************************************************************************/
/*                       Reset Markers Topo Helper                            */
/******************************************************************************/
// Resets the marker variable in all the nodes of the network rooted at 'nd'
// specified when called.  NOTE:  this routine follows the topology, not
// the NodeArray.  This means that only a subtree can be reset by providing
// the subtree's root, however this can be costly time wise as this logic allows
// for traversing paths more than once!  We cannot end up in an infinite loop
// though because our network has directed edges so we are safe (if albeit
// inefficient).
//
void ResetMarkersTopoHelper(NODE nd)
{
  NODE tmp_nd;

  nd->marker = 0;

  if (nd->first_child_node != NULL)
  {
    tmp_nd = nd->first_child_node;

    if (tmp_nd->reticulate == 1)
    {
      hyb_parent_hit_ctr++;

      if (tmp_nd->parent1_node == nd)
        last_hyb_parent_hit = 1;
      else if (tmp_nd->parent2_node == nd)
        last_hyb_parent_hit = 2;
      else
      {
        printf("NR ERROR: problem with ResetMarkersTopoHelper routine -- program aborted\n");
        exit(0);
      }
    }

    while(tmp_nd != NULL)
    {
      ResetMarkersTopoHelper(tmp_nd);

      // if dealing with reticulate node need to make sure we get the
      // next sibling node correctly
      if (tmp_nd->reticulate == 1)
      {
        if (last_hyb_parent_hit == 1)
	{
	  tmp_nd = tmp_nd->next_p1_sib;
	}
	else if (last_hyb_parent_hit == 2)
	{
	  tmp_nd = tmp_nd->next_p2_sib;
	}
	else
	{
	  printf("NR ERROR: parent hit problem in ResetMarkersTopoHelper -- program aborted\n");
	  exit(0);
	}
      }
      else // otherwise deal as reg sib
      {
        tmp_nd = tmp_nd->next_sib_node;
      }
    }
  }
  else
  {
    return;
  }

}

/******************************************************************************/
/*                            ResetWriteMarkers                               */
/******************************************************************************/
void ResetWriteMarkers(NETWORK nt)
{
  // this routine resets the write markers -- it goes through the node
  // array to do this so there is no topology aspect
 
  int node_i;

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    nt->NodeArray[node_i]->marked_to_write = 0;
  }

  return;
}

/******************************************************************************/
/*                           Set Markers Topo                                 */
/******************************************************************************/
void SetMarkersTopo(NODE nd, int val)
{
  hyb_parent_hit_ctr = 0;

  SetMarkersTopoHelper(nd, val);

  return;
}


/******************************************************************************/
/*                        Set Markers Topo Helper                             */
/******************************************************************************/
void SetMarkersTopoHelper(NODE nd, int val)
{
  // This routine sets marker variable for all nodes at and below the provided
  // node.  THis follows a topo order, NOT the node array.  Logic does allow
  // for traversing more than once but no risk of infinite loop due to our
  // directed edges.
  
  NODE tmp_nd;

  nd->marker = val;

  if (nd->first_child_node != NULL)
  {
    tmp_nd = nd->first_child_node;

    if (tmp_nd->reticulate == 1)
    {
      hyb_parent_hit_ctr++;

      if (tmp_nd->parent1_node == nd)
        last_hyb_parent_hit = 1;
      else if (tmp_nd->parent2_node == nd)
        last_hyb_parent_hit = 2;
      else
      {
        printf("NR ERROR: problem with last parent hit in SetMarkersTopoHelper -- program aborted\n");
        exit(0);
      }
    }

    while(tmp_nd != NULL)
    {
      SetMarkersTopoHelper(tmp_nd, val);

      // if dealing with reticulate node need to make sure we get the
      // next sibling node correctly
      if (tmp_nd->reticulate == 1)
      {
        if (last_hyb_parent_hit == 1)
	{
	  tmp_nd = tmp_nd->next_p1_sib;
	}
	else if (last_hyb_parent_hit == 2)
	{
	  tmp_nd = tmp_nd->next_p2_sib;
	}
	else
	{
	  printf("NR ERROR: hit counter problem in SetMarkersTopoHelper -- program aborted\n");
	  exit(0);
	}
      }
      else
      {
        tmp_nd = tmp_nd->next_sib_node;
      }
    }
  }
  else
  {
    return;
  }

  return;
}

/******************************************************************************/
/*                        Tree Hunting Related Functions                      */
/******************************************************************************/
NODE FindNode(NETWORK nt, int internal_id)
{
  int nd_i;
  NODE found_node;

  found_node = NULL;

  for (nd_i = 0; nd_i < nt->num_unique_nodes; nd_i++)
  {
    if (nt->NodeArray[nd_i]->internal_node_id == internal_id)
    {
      found_node = nt->NodeArray[nd_i];
      break;
    }
  }

  if (found_node == NULL)
  {
    printf("NR ERROR: Could not find node with internal id %d in the node array -- this should not have occured\n", internal_id);
    exit(0);
  }

  return found_node;
}

/******************************************************************************/
/*                         Max Int (of 4 integers)                            */
/******************************************************************************/
int MaxInt(int a, int b, int c, int d)
{
  if ( (a >= b) &&
       (a >= c) &&
       (a >= d) )
  {
    return a;
  }
  else if ( (b >= a) &&
            (b >= c) &&
            (b >= d) )
  {
    return b;
  }
  else if ( (c >= a) &&
            (c >= b) &&
            (c >= d) )
  {
    return c;
  }
  else if ( (d >= a) &&
            (d >= b) &&
            (d >= c) )
  {
    return d;
  }
  else
  {
    printf("NR ERROR:  unexpected problem in max int -- program aborted\n");
    exit(0);
  }

}

/******************************************************************************/
/*                          CalcParsimonyScore                                */
/******************************************************************************/
void CalcParsimonyScore(NETWORK nt)
{
  int hd_sum_strand, total_hd_sum;
  int node_i, hchromo_i, strand_i;

  total_hd_sum = 0;

  // start by doing all the work for the root
  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      hd_sum_strand = 0;
      ResetMarkersAll(nt);

      hd_sum_strand = CalcStrandHD(nt->root, hchromo_i, strand_i);

      //printf("NR DEBUG: hd sum strand for root of hchromo %d strand %d is %d\n", hchromo_i, strand_i, hd_sum_strand);

      total_hd_sum = total_hd_sum + hd_sum_strand;
    }
  }

  // now repeat the process for all reticualte nodes -- but be careful in that the
  // first call is done with the reticulate node as the parent not the child -- there
  // is no change from parent to hybrid node, so don't worry about that either
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->reticulate != 0)
    {
      for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
      {
        for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
        {
          hd_sum_strand = 0;
          ResetMarkersAll(nt);

          hd_sum_strand = CalcStrandHD(nt->NodeArray[node_i], hchromo_i, strand_i);
          //printf("NR DEBUG: hd sum strand for retic of hchromo %d strand %d is %d\n", hchromo_i, strand_i, hd_sum_strand);

          total_hd_sum = total_hd_sum + hd_sum_strand;
        }
      }
    }
  }

  // now assign the final value
  nt->parsimony_score = total_hd_sum;
  
}

/******************************************************************************/
/*                          CalcStrandHD                                      */
/******************************************************************************/
int CalcStrandHD(NODE rt, int hchromo_i, int strand_i)
{
  int orig_call;

  // This routine starts at the given node and figures out how many changes
  // occur for the given strand from the root down to where it shows
  // up in extant taxa or hits a reticulate node.  This routine is called
  // for the real root node and any reticulate node which acts as a root
  // in this case.

  orig_call = 1;
  strand_summed_hd = 0;

  if (rt->reticulate != 0)
  {
    // if this is a reticulate subroot, the retic node is the parent
    HDDFS(rt, rt->first_child_node, hchromo_i, strand_i, orig_call);
  }
  else 
  {
    // with the real root, there is no parent to pass
    HDDFS(NULL, rt, hchromo_i, strand_i, orig_call);
  }
  
  // this variable is a global which is "cheating" in the world of
  // recursion, but it sure makes our life easier, so, so be it!
  return strand_summed_hd;

}

/******************************************************************************/
/*                    HDDFS (hamming distance depth first search)             */
/******************************************************************************/
void HDDFS(NODE parent, NODE child, int hchromo_i, int strand_i, int orig_call)
{
  NODE tmp_nd;
  int hd;

  // this is the recursive work....

  if ( ((parent != NULL) && (parent->reticulate != 0) && (orig_call != 1)) ||
       ((child != NULL) && (child->reticulate != 0)) )
   
  {
    // DEUBBING
    //printf("NR DEBUG:  not proceeding down this lineage in HDDFS as retic node hit\n");
    
    return;
  }
  else 
  {
    // do the work
    if ( (child != NULL) && (child->marker == 0) && (parent != NULL) )
    {
      hd = HammingDistanceStrings(parent->hom_chromos[hchromo_i]->strands[strand_i]->sequence, child->hom_chromos[hchromo_i]->strands[strand_i]->sequence);

      child->marker = 1;

      // update our global variable
      strand_summed_hd = hd + strand_summed_hd;
    }
  }

  // deal with the traversal
  if ( (child != NULL) && (child->first_child_node != NULL) )
  {
    HDDFS(child, child->first_child_node, hchromo_i, strand_i, 0);

    // do the sib traversal
    if (child->first_child_node->reticulate != 0)
    {
      if (child->first_child_node->parent1_node == child)
        tmp_nd = child->first_child_node->next_p1_sib;
      else if (child->first_child_node->parent2_node == child)
        tmp_nd = child->first_child_node->next_p2_sib;
      else
      {
        printf("NR ERROR:  Unexpected trouble with parsimony score calc -- program aborted\n");
	exit(0);
      }
    }
    else
    {
      tmp_nd = child->first_child_node->next_sib_node;
    }

    if (tmp_nd != NULL)
    {
      HDDFS(child, tmp_nd, hchromo_i, strand_i, 0);
      tmp_nd = tmp_nd->next_sib_node;
    }
  }

  return;
}
