/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#ifndef _NETWORK_H
#define _NETWORK_H

/**************************************************************************/
/*                          Network Structure                             */
/**************************************************************************/
struct network {
  NODE root;
  NODE outgroup;
  NODE subtree_roots[MAX_NUM_HYBRIDS]; // indexed by subtree, internal id stored
  NODE outgroups_for_subtrees[MAX_NUM_HYBRIDS]; // indexed by subtree, 
                                               // stores external id
  NODE NodeArray[MAX_NODES];        // has all nodes 
  NODE parental_subtree_root;
  NODE hybrid_subtree_root;
				    
  // note:  a node's id should be the same as its index into NodeArray
  // note:  an extant leaf's id should be its index into ExtantLeafArray
  
  int node_id_ctr;          // tracks what the next node id should be
  int num_unique_nodes; 
  int num_unique_extant_leaves;
  int num_reticulate_nodes;
  int num_unique_reticulate_nodes;
  int parsimony_score;

  int outgroup_external_id; 
  int outgroup_ids_for_subtrees[MAX_NUM_HYBRIDS]; // indexed by subtree, 
                                                  // stores external id
};

typedef struct network NETWORKTYPE;
typedef struct network* NETWORK;

// set-up/clean-up
NETWORK NewNetwork();
void CleanUpMemory(NETWORK nt);

// Parsing Related Functions
void ParseNewick(NETWORK nt);
int GetToken();

// Assignment Functions
void DownwardAssignRound(NODE nd, int base_index, int hchromo_i, int strand_i);
void UpwardOptionRound(NODE nd, int base_index, int hchromo_i, int strand_i);

// Assingment Helper Functions
void PopulateExtantBaseValues(NODE nd);
void SetAllPosBaseValues(NODE p_node, NODE c_node, int base_index, 
                         int hchromo_i, int strand_i);
int CheckForEmptyDescInter(NODE nd, int base_index, int hchromo_i, int strand_i);
void DoDescUnion(NODE nd, int base_index, int hchromo_i, int strand_i);
void DoDescIntersection(NODE nd, int base_index, int hchromo_i, int strand_i);
void ResetAllBaseValues(NODE nd, int base_index, int hchromo_i, int strand_i);
void SetSingleBaseValue(NODE nd, int base_index, int hchromo_i, int strand_i,
                        int chosen_base);
void SetBaseChar(NODE nd, int base_index, int hchromo_i, int strand_i, 
                 int chosen_base);
int ReturnDesiredBase(NODE nd, int base_index, int hchromo_i, int strand_i);
int CountKids(NODE nd);
int ReturnMajorityBase(NODE nd, NODE parent, int base_index, int hchromo_i, 
                       int strand_i);

// Helper Functions 
int CountExtantImpactedByHybrid(NETWORK nt);
int CountExtantImpactedByParents(NETWORK nt);
int CountExtantImpactedByParent(NETWORK nt, int paffval);
int CountExtantNotImpacted(NETWORK nt);
int IsTree(NETWORK nt);
void ResetMarkersAll(NETWORK nt);
void ResetMarkersTopo(NODE nd); 
void ResetMarkersTopoHelper(NODE nd);
void ResetWriteMarkers(NETWORK nt);
void SetMarkersTopo(NODE nd, int val);
void SetMarkersTopoHelper(NODE nd, int val);
NODE FindNode(NETWORK nt, int internal_id);
int MaxInt(int a, int b, int c, int d);

// score functions
void CalcParsimonyScore(NETWORK nt);
int  CalcStrandHD(NODE rt, int hchromo_i, int strand_i);
void HDDFS(NODE parent, NODE child, int hchromo_i, int hchromo_j, int orig_call);


// NOTE:  topo based routines are hard coded for 1 hybrid -- this can be 
//        overcome using another marker for the node and having that be of use
//        when deciding if going to p1next or p2next -- see NM code
//        for details -- esp the DFS routine -- note that the data
//        structure in NM changed drastically because of retic siblings
//        which are not remotely possible here!!!
//
//
// NOTE!!!!!!!!!   because we are limited to one hybrid for this code and 
//                 the way we have the next sib stuff set up and most 
//                 importantly how we create the parents of the hybrid
//                 and hook them to the hybrid, we are guaranteed that the
//                 hybrid node is the first child of its p1 and p2 -- this
//                 assumption is use ALOT in processing!!!!!!!!!!

#endif
