/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<limits.h>
#include<float.h>
#include"globals.h"
#include"node.h"


/******************************************************************************/
/*                            NewSingleStrand                                 */
/******************************************************************************/
SINGLESTRAND NewSingleStrand(int strand_id) 
{
  char *temp_ptr;
  char strand_id_str[10];
  SINGLESTRAND strand;
  int seq_i;

  strand = (SINGLESTRAND)malloc(sizeof(SINGLESTRANDTYPE));

  if (strand == NULL)
  {
    printf("NR ERROR: Could not allocate space for strand\n");
    exit(0);
  }

  // doing + 1 here for \0 terminator
  temp_ptr = (char *) calloc((sequence_length + 1),sizeof(char));

  if (temp_ptr == NULL)
  {
    printf("NR ERROR: Cannot allocate memory for a strand's sequence\n");
    printf("NR ERROR: a sequence length of %d was requested\n", sequence_length);
    exit(0);
  }

  strand->sequence = temp_ptr;

  // if the sequence lenght is long enough put in a temp seq
  if (sequence_length > 25)
  {
    strcpy(strand->sequence, "init strand");
    sprintf(strand_id_str, "%d", strand_id);
    strcat(strand->sequence, strand_id_str);
  }

  // initialize the base option array which gets used for sequence assignment
  for (seq_i = 0; seq_i < MAX_SEQ_LENGTH; seq_i++)
  {
    strand->base_option_array[0][seq_i] = 0;   
    strand->base_option_array[1][seq_i] = 0;   
    strand->base_option_array[2][seq_i] = 0;   
    strand->base_option_array[3][seq_i] = 0;   
  }

  strand->number = strand_id;
  strand->ancestor_node_id = -1;
  strand->id = strand_id;
  strand->fixed_seq = 0;

  return strand;
}

/******************************************************************************/
/*                            NewHomChrom                                     */
/******************************************************************************/
HCHROMOSOME NewHomChrom(int hchromo_id) 
{
  int strand_i;
  HCHROMOSOME hchromo;

  hchromo = (HCHROMOSOME)malloc(sizeof(HCHROMOSOMETYPE));

  if (hchromo == NULL)
  {
    printf("NR ERROR: Could not allocate space for hchromosome\n");
    exit(0);
  }

  for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
  {
    hchromo->strands[strand_i] = NewSingleStrand(strand_i);
  }

  hchromo->number = hchromo_id;
  hchromo->id = hchromo_id;

  return hchromo;
}

/******************************************************************************/
/*                                 NewNode                                    */
/******************************************************************************/
NODE NewNode(int nd_id) 
{
  NODE nd;
  
  int hchromo_i, i;

  // make sure this node will not exceed our limit
  if (nd_id >= (MAX_NODES))
  {
    printf("NR ERROR: Simulation could not proceed as MAX_NODES (%d) is exceeded\n", MAX_NODES);
    exit(0);
  }

  // make the new node
  nd = (NODE)malloc(sizeof(NODETYPE));

  if (nd == NULL)
  {
    printf("NR ERROR: Could not allocate space for node\n");
    exit(0);
  }

  // use the values we were given
  nd->internal_node_id = nd_id;

  // set all others to default values
  nd->external_node_id = -1;
  nd->first_child_node = NULL;
  nd->prev_sib_node = NULL;
  nd->next_sib_node = NULL;
  nd->parent1_node = NULL;
  nd->parent2_node = NULL;
  nd->next_p1_sib = NULL;
  nd->next_p2_sib = NULL;
  nd->extant_leaf = 0;   
  nd->reticulate = 0;
  nd->temporary = 0;
  nd->parental_aff = PARENTAL_AFF_DEFAULT;
  nd->brlen1 = 0.0;
  nd->brlen2 = 0.0;
  nd->marker = 0;
  nd->marked_to_write = 0;

  // set up the homolgous chromosomes/strands
  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    nd->hom_chromos[hchromo_i] = NewHomChrom(hchromo_i);
  }

  // initialize the impacted_by arrays to -1s
  for (i = 0; i < MAX_NUM_HYBRIDS; i++)
  {
    nd->impacted_by_hybrid[i] = -1;
    nd->impacted_by_parents[i] = -1;
  }

  return nd;

}  // end of NewNode()

/******************************************************************************/
/*                         HammingDistanceTotal                               */
/******************************************************************************/
int HammingDistanceTotal(NODE nd_1, NODE nd_2)
{
  // here we define that hamming distance between two nodes does
  // not care about the number of homologous chromosomes/strands -- simply
  // the number of sequences to compare (so if one node has one
  // homologous chromosome with two strands and the other node has two 
  // homologous chromosomes each with one strand, that is fine)
  //
  // here, if the number of sequences do not match, the number of 
  // nucleotides which are differing is added to the hamming distance
  // (e.g. if one node has two sequences of 10 each and the other
  // node has 1 sequence of 5 then the starting hamming distance
  // is 10 (as there is one extra string of length 10) -- the difference
  // in string length is captured and added to the hamming distance in
  // the HammingDistanceStrings function below

  int num_seqs_1;
  int num_seqs_2;
  int num_seqs_to_compare;

  int hd;
  int seq_pair_i;

  int nd_1_hchromo_index;
  int nd_1_strand_index;
  int nd_2_hchromo_index;
  int nd_2_strand_index;

  char str_1[sequence_length];
  char str_2[sequence_length];

  hd = 0;
  num_seqs_1 = num_hom_chromos * num_strands_per_hom_chrom;
  num_seqs_2 = num_hom_chromos * num_strands_per_hom_chrom;

  if (num_seqs_1 == num_seqs_2)
  {
    num_seqs_to_compare = num_seqs_1;
  }
  else if (num_seqs_1 > num_seqs_2)
  {
    num_seqs_to_compare = num_seqs_2;
    hd = (num_seqs_1 - num_seqs_2) * sequence_length;
  }
  else
  {
    num_seqs_to_compare = num_seqs_1;
    hd = (num_seqs_2 - num_seqs_1) * sequence_length;
  }

  // DEBUGGING
  //printf("NR DEBUG:  hd before string comp is %d\n", hd);

  // now call HammingDistanceStrings for each pair of strings
  // we have available for comparing
  nd_1_hchromo_index = 0;
  nd_1_strand_index = 0;
  nd_2_hchromo_index = 0;
  nd_2_strand_index = 0;

  for (seq_pair_i = 0; seq_pair_i < num_seqs_to_compare; seq_pair_i++)
  {
    // figure out which strings we will be comparing and copy them into 
    // str_1 and str_2 -- note that since the number of homolgous
    // chromosomes/strands may not match, need to track the index of 
    // the next string for each node and basically do mod arithmetic by 
    // hand -- there is probably a more elegant way to do this, but this 
    // should be straightforward and work
    
    strcpy(str_1, nd_1->hom_chromos[nd_1_hchromo_index]->strands[nd_1_strand_index]->sequence);

    nd_1_strand_index++;
    if (nd_1_strand_index == num_strands_per_hom_chrom)
    {
      nd_1_hchromo_index++;
      nd_1_strand_index = 0;
    }

    strcpy(str_2, nd_2->hom_chromos[nd_2_hchromo_index]->strands[nd_2_strand_index]->sequence);

    nd_2_strand_index++;
    if (nd_2_strand_index == num_strands_per_hom_chrom)
    {
      nd_2_hchromo_index++;
      nd_2_strand_index = 0;
    }

    hd = HammingDistanceStrings(str_1, str_2) + hd;
    
  }

  return hd;
}

/******************************************************************************/
/*                           HammingDistanceStrings                           */
/******************************************************************************/
int HammingDistanceStrings(char * str_1, char * str_2)
{
  int char_i, length, count;

  // if the strings differ in length we automatically have a hamming
  // distance > 0; so add that to count right away and then we can 
  // stop comparing strings when the shortest string is exhausted 

  length = MinInt(strlen(str_1), strlen(str_2));
  count = 0;
  count = count + (abs(strlen(str_1) - strlen(str_2))); 

  for (char_i = 0; char_i < length; char_i++)
  {
    if (str_1[char_i] != str_2[char_i])
      count++;
  }

  return count;
}

/******************************************************************************/
/*                                 MinInt                                     */
/******************************************************************************/
int MinInt(int a, int b)
{
  if (a < b)
    return a;
  else
    return b;
}
