/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#ifndef _NODE_H
#define _NODE_H

/**********************************************************************/
/*       A single strand is one string of nucleotides                 */
/**********************************************************************/
struct single_strand {
  char* sequence;
  int   number; // this is the strand number biologically -- not the index!
  int   id;     // this is the index of the strand for processing purposes
  int   ancestor_node_id;
  int   fixed_seq; // 1 means sequence is fixed, 0 is open
  int   base_option_array [4][MAX_SEQ_LENGTH]; // simplistic implementation
};

typedef struct single_strand SINGLESTRANDTYPE;
typedef struct single_strand* SINGLESTRAND;


/**********************************************************************/
/*       A homologous chromosome is a grouping of strands             */
/**********************************************************************/
struct homologous_chromosome {
  SINGLESTRAND strands[MAX_NUM_STRANDS_PER_HC];
  int number;  // this is the homologous chromosome number biologically 
               // not the index!
  int id;      // this is the index of the homologous chromosome for 
               // processing purposes
};

typedef struct homologous_chromosome HCHROMOSOMETYPE;
typedef struct homologous_chromosome* HCHROMOSOME;



/**********************************************************************/
/*    A node is the primary object of our network(s).                 */
/*                                                                    */
/*    By starting at the network's root node, one can follow the      */
/*    node links and traverse the network.                            */
/*                                                                    */
/*    The branch lengths (brlen1 and brlen2) associated with a        */
/*    node are the edges which terminate at the given node.           */
/*    Hence brlen1 refers to the branch length between the node and   */
/*    its parent1 (and likewise for brlen2 and parent2 if present).   */
/*                                                                    */
/*    Note:  the root does not have any parents by definition and     */
/**********************************************************************/

struct node {
  struct node *first_child_node, *prev_sib_node, *next_sib_node, *parent1_node, *parent2_node, *next_p1_sib, *next_p2_sib;

  int internal_node_id;   // this is the index into the node array 
  int external_node_id;
  int extant_leaf;        // 0 false, 1 true
  int reticulate;         // 0: tree node 1+: newtork node (1=hybrid, 2=lgt)
  int temporary;          // 0 if node part of final soln, 1 if temp placeholder
  int impacted_by_hybrid[MAX_NUM_HYBRIDS]; // -1 default means not impacted, 
                                           // otherwise value indicates
                                           // the subtree/hybrid index
  int impacted_by_parents[MAX_NUM_HYBRIDS]; // -1 default means not impacted 
                                            // otherwise value indicates
                                            // the subtree/hybrid index
  int parental_aff; // which parent of the single hybrid with
                    // which the node is affiliated either -2 or -3 (p1, p2)

  double brlen1, brlen2;  // incoming branch lengths to this node

  HCHROMOSOME hom_chromos[MAX_NUM_HOM_CHROMOS];

  int marker;             // marker variable for generic use
  int marked_to_write;    // marker to indicate we are going to be writing
                          // this node into the input file

};
typedef struct node NODETYPE;
typedef struct node* NODE;

// the "constructor"
SINGLESTRAND NewSingleStrand(int strand_id);
HCHROMOSOME NewHomChrom(int hchromo_id);
NODE    NewNode(int nd_id);

// utility functions
int HammingDistanceTotal(NODE nd1, NODE nd2);
int HammingDistanceStrings(char * str_1, char * str_2);
int MinInt(int a, int b);

#endif
