/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<float.h>
#include<limits.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"phylipinter.h"
#include"randipinter.h"
#include"output.h"
#include"twister.h"
#include"parsub.h"


/******************************************************************************/
/*                           ReconParentalOffspring                           */
/******************************************************************************/
void ReconParentalSubtree(NETWORK nt)
{
  int psub_i, psub_num;

  // FOR DEBUGGING
  //NODE nd;
  //int node_i;

  // need to identify which extant taxa are parental offspring
  IdentifyParentalOffspring(nt);

  // need to make sure our sets are okay in general
  CheckSubsets(nt);

  // DEBUGGING
  //printf("NR DEBUG: from recon parental subtree we have %d node(s) to be impacted by parents\n", num_impacted_by_parents);
  //for(node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  //{
    //nd = nt->NodeArray[node_i];

    //if (nd->impacted_by_parents[0] != -1)
    //{
      //printf("NR DEBUG:  node external id %d (internal id %d) is impacted by parents, with affliation of %d\n", nd->external_node_id, nd->internal_node_id, nd->parental_aff);
    //}
  //}
  // END DEBUGGING

  // now set the number parental subtrees we will be dealing with depending
  // on whether the extreme custom (2) or any other method (1) is requested
  psub_num = 1;

  if (parental_offspring_id_method == 3)
    psub_num = 2;

  for (psub_i = 0; psub_i < psub_num; psub_i++)
  {
    // need to do the work of reconstructing the topology
    ReconParentalTopology(nt, psub_i);

    // read the results if PHYLIP or RANDIP was called
    if ( (parental_offspring_id_method != 3) && (num_impacted_by_parents > 0) )
    {
      ReadParentalSubtreeTopology(nt);
    }
    else if ( (parental_offspring_id_method == 3) && 
              (psub_i == 0) && 
	      (num_impacted_by_p1 > 0) )
    {
      ReadParentalSubtreeTopology(nt);
    }
    else if ( (parental_offspring_id_method == 3) && 
              (psub_i == 1) && 
	      (num_impacted_by_p2 > 0) )
    {
      ReadParentalSubtreeTopology(nt);
    }

    // make sure we assign the parental subtree root and its attributes
    nt->parental_subtree_root = nt->root;
    nt->parental_subtree_root->temporary = 1;

    // unhook the outgroup so we don't get confused with it
    // this only needs to be done if PHYLIP or RANDIP was called
    if ( (parental_offspring_id_method != 3) && (num_impacted_by_parents > 0) )
    {
      UnhookParentalSubtreeOutgroup(nt);
    }
    else if ( (parental_offspring_id_method == 3) &&
              (psub_i == 0) &&
	      (num_impacted_by_p1 > 0) )
    {
      UnhookParentalSubtreeOutgroup(nt);
    }
    else if ( (parental_offspring_id_method == 3) &&
              (psub_i == 1) &&
	      (num_impacted_by_p2 > 0) )
    {
      UnhookParentalSubtreeOutgroup(nt);
    }
  
    // need to do parental assignment if appropriate value was above 0
    if ( (parental_offspring_id_method != 3) && (num_impacted_by_parents > 0) )
    {
      AssignParentalLineages(nt);
    }
    else if ( (parental_offspring_id_method == 3) && 
              (psub_i == 0) &&
	      (num_impacted_by_p1 > 0) )
    {
      AssignParentalLineagesExtreme(nt, psub_i);
    }
    else if ( (parental_offspring_id_method == 3) && 
              (psub_i == 1) &&
	      (num_impacted_by_p2 > 0) )
    {
      AssignParentalLineagesExtreme(nt, psub_i);
    }

    // need to get parental offspring hooked into subtree
    if ( (parental_offspring_id_method != 3) && (num_impacted_by_parents > 0) )
    {
      ConnectSubtrees(nt);
    }
    else if ( (parental_offspring_id_method == 3) &&
              (psub_i == 0) &&
	      (num_impacted_by_p1 > 0) )
    {
      ConnectSubtrees(nt);
    }
    else if ( (parental_offspring_id_method == 3) &&
              (psub_i == 1) &&
	      (num_impacted_by_p2 > 0) )
    {
      ConnectSubtrees(nt);
    }
             
    // now that we are all done with this we can unhook the root
    // from the parental subtree and reset any markers
    nt->parental_subtree_root->first_child_node = NULL;

    ResetMarkersAll(nt);

  } // next parental subtree if doing extreme custom


  // DEBUGGING -- only works in certain cases
  //printf("NR DEBUG: parent1 is %d parent2 is %d  and the offspring of the hybrid is %d \n", nt->hybrid_subtree_root->parent1_node->internal_node_id, nt->hybrid_subtree_root->parent2_node->internal_node_id, nt->hybrid_subtree_root->first_child_node->internal_node_id);
  
  // DEBUGGING -- only works in certain cases
  // printf("NR DEBUG: parent1 is %d parent2 is %d  and the offspring of the hybrid is %d  and the offspring for p1 is % d (and it thinks its parent is %d) and for p2 is %d (and it thinks its parent is %d) \n", nt->hybrid_subtree_root->parent1_node->internal_node_id, nt->hybrid_subtree_root->parent2_node->internal_node_id, nt->hybrid_subtree_root->first_child_node->internal_node_id, nt->hybrid_subtree_root->parent1_node->first_child_node->next_p1_sib->internal_node_id, nt->hybrid_subtree_root->parent1_node->first_child_node->next_p1_sib->parent1_node->internal_node_id, nt->hybrid_subtree_root->parent2_node->first_child_node->next_p2_sib->internal_node_id, nt->hybrid_subtree_root->parent2_node->first_child_node->next_p2_sib->parent1_node->internal_node_id);
 
  // DEBUGGING -- only works in certain cases
  //printf("NR DEBUG: parent1 is %d parent2 is %d  p1's first child is %d and the next is %d and the offspring of the hybrid is %d \n", nt->hybrid_subtree_root->parent1_node->internal_node_id, nt->hybrid_subtree_root->parent2_node->internal_node_id, nt->hybrid_subtree_root->parent1_node->first_child_node->internal_node_id, nt->hybrid_subtree_root->parent1_node->first_child_node->next_p1_sib->internal_node_id, nt->hybrid_subtree_root->first_child_node->internal_node_id);


  return;
} 

/******************************************************************************/
/*                           IdentifyParentalOffspring                        */
/******************************************************************************/
void IdentifyParentalOffspring(NETWORK nt)
{
  // if we are doing one of the customs, we already know which extant taxa
  // are the ones impacted by the parent(s); otherwise we need to
  // create this set
  
  if ( (parental_offspring_id_method != 2) &&
       (parental_offspring_id_method != 3) )
  {
    if (parental_offspring_id_method == 0)
    {
      ClosestNeighborSelection(nt);
    }
    else if (parental_offspring_id_method == 1)
    {
      Closest2xHybridSelection(nt);
    }
    else
    {
      printf("NR ERROR:  Invalid method for selecting parental offspring taxa -- program aborted\n");
      exit(0);
    }
  }
}

/******************************************************************************/
/*                          ClosestNeighborSelection                          */
/******************************************************************************/
void ClosestNeighborSelection(NETWORK nt)
{
  int node_i, node_j, tie_yn, min_hd, min_hd_index, y, z;
  int hd[MAX_NODES];
  double rand_val;
  NODE nd;
  NODE comp_nd;


  // WARNING:  remember this is hard-coded for one hybrid subtree!

  for(node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    nd = nt->NodeArray[node_i];

    // only worry about extant leaves that are NOT impacted by the hybrid
    // and if it is NOT the outgroup for the network!
    if ( (nd->extant_leaf == 1) && 
         (nd->impacted_by_hybrid[0] == -1) &&
	 (nt->outgroup != nd) )
    {

      // DEBUGGING
      //printf("NR DEBUG: processing node %d\n", nt->NodeArray[node_i]->internal_node_id);
     
      // first initialize min_hd and the hd array for this round
      min_hd = INT_MAX;

      for (node_j = 0 ; node_j < nt->num_unique_nodes; node_j++)
      {
        hd[node_j] = -1;
      }

      // go through and calculate the average hamming distance between
      // this node and all the other extant taxa and look for the min
      for (node_j = 0; node_j < nt->num_unique_nodes; node_j++)
      {
        comp_nd = nt->NodeArray[node_j];

        if ( (nd != comp_nd) && (comp_nd->extant_leaf == 1) )
	{
          hd[node_j] = HammingDistanceTotal(nd, nt->NodeArray[node_j]);

	  // DEBUGGING
	  //printf("NR DEBUG: HD between %d and %d is %d\n", nt->NodeArray[node_i]->internal_node_id, nt->NodeArray[node_j]->internal_node_id, hd[node_j]);
	
	  if (hd[node_j] < min_hd)
	  {
	    min_hd = hd[node_j];
	    min_hd_index = node_j;
	    tie_yn = 0;
	  }
	  else if (hd[node_j] == min_hd)
	  {
	    tie_yn = 1;
	  }
	}
      }

      if (tie_yn == 0)
      {
        // DEBUGGING
	//printf("NR DEBUG: no tie!\n");

        // see if we want the node to be a parental offspring based
	// on if its neighbor is impacted by the hybrid or not
        if (nt->NodeArray[min_hd_index]->impacted_by_hybrid[0] != -1) 
	{
	  nd->impacted_by_parents[0] = 0;

	  // DEBUGGING
	  //printf("NR DEBUG: we have a node to be impacted by parents\n");
	}
	else // not impacted
	{
	  nd->impacted_by_parents[0] = -1;

	  // DEBUGGING
	  //printf("NR DEBUG: we have a node NOT impacted by parents\n");
	}
      }
      else // there is a tie
      {
        // DEBUGGING
	//printf("NR DEBUG: tie!\n");

        // initialize local variables which will count how many
	// of the min hd nodes are impacted or not impacted by
	// the hybrid node
	y = 0;
	z = 0;
        
	for (node_j = 0; node_j < nt->num_unique_nodes; node_j++)
	{
	  if (hd[node_j] == min_hd)
	  {
	    if (nt->NodeArray[node_j]->impacted_by_hybrid[0] != -1)
	      y++;
	    else if (nt->NodeArray[node_j]->impacted_by_hybrid[0] == -1)
	      z++;
	  }
	}

	if (y > z)
	{
	  // DEBUGGING
	  //printf("NR DEBUG: parental impact after tie examination\n");

	  // more min hd neighbors are impacted so make this
	  // node a parental offspring taxon
	  nd->impacted_by_parents[0] = 0;
	}
	else if (y < z)
	{
	  // DEBUGGING
	  //printf("NR DEBUG: no parental impact after tie examination\n");

	  nd->impacted_by_parents[0] = -1;
	}
	else // we have a true tie
	{
	  // DEBUGGING
	  //printf("NR DEBUG: true true tie\n");

	  // flip a coin to decide
	  rand_val = genrand_real1();

	  if (rand_val < .5)
	    nd->impacted_by_parents[0] = 0;
	  else
	    nd->impacted_by_parents[0] = -1;
	}
      }
    } // extant leaf and not hybrid impacted check
  } // next node

  return;
}

/******************************************************************************/
/*                          Closest2xHybridSelection                          */
/******************************************************************************/
void Closest2xHybridSelection(NETWORK nt)
{
  NODE nd;

  int hi_ctr;
  int node_i;
  int num_non_imp_extant_leaves;
  int desired_parental_count;
  int desired_parental_ctr;
  int max_hd, min_hd, current_hd, last_used_hd;
  int num_to_unpick, which_to_unpick, which_ctr;
  int num_tries, max_num_tries, num_with_current_hd, num_found_this_round;

  int hd_from_hybrid[MAX_NODES];
  int picked[MAX_NODES];

  // WARNING:  remember this is hard coded for one hybrid subtree!
  
  // first we need to count how many extant taxa were identified on
  // input as being impacted by the hybrid 
 
  hi_ctr = 0;

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->impacted_by_hybrid[0] != -1)
      hi_ctr++;
  }

  // DEBUGGING
  //printf("NR DEBUG: hi ctr %d\n", hi_ctr);

  // don't forget the outgroup is NOT eligible so do a -1 here
  num_non_imp_extant_leaves = nt->num_unique_extant_leaves - hi_ctr - 1;
  desired_parental_count = 2 * hi_ctr;

  // DEBUGGING
  //printf("NR DEBUG: num-non-impacted-extant-leaves ctr %d and desired_parental_count %d\n",num_non_imp_extant_leaves, desired_parental_count);

  if ( (num_non_imp_extant_leaves - desired_parental_count) < 0)
  {
    printf("NR ERROR:  There are not enough extant leaves which are not impacted by the hybrid to choose the parental offspring taxa using the closest 2x hybrid method -- choose another method or change impacted/non-impacted input -- program aborted with value of 99\n");
    exit(99);
  }
  else
  {
    // need to pick the 2x nodes closest to the hybrid node

    // initialize arrays
    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      hd_from_hybrid[node_i] = -1;
      picked[node_i] = 0;
    }

    max_hd = -1;
    min_hd = INT_MAX;
    current_hd = INT_MAX;
    
    // first set up hd from hybrid for all extant leaves not
    // impacted by the hybrid and not the outgroup for the
    // network itsel (these are the ones eligible for being 
    // parental offspring)
    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      nd = nt->NodeArray[node_i];

      if ( (nd->extant_leaf == 1) && 
           (nd->impacted_by_hybrid[0] == -1) &&
	   (nt->outgroup != nd) )
      {
        hd_from_hybrid[node_i] = HammingDistanceTotal(nd, nt->root); 

        // DEBUGGING
        //printf("NR DEBUG: hd from hybrid for node %d is    %d\n", node_i, hd_from_hybrid[node_i]);

	if (hd_from_hybrid[node_i] < min_hd)
	{
	  min_hd = hd_from_hybrid[node_i];
	}
	if (hd_from_hybrid[node_i] > max_hd)
	{
	  max_hd = hd_from_hybrid[node_i];
	}
      }
    }

    // now find our 2x nodes and label them as impacted by parent 
    desired_parental_ctr = 0;
    current_hd = min_hd;
    last_used_hd = min_hd;

    // DEBUGGING
    //printf("NR DEBUG: after hd calc min hd is %d and max hd is %d  and the current hd is %d\n", min_hd, max_hd, current_hd);

    while ( (desired_parental_ctr < desired_parental_count) &&
            (current_hd <= max_hd) )
    {
      // see if we have an extant leaf with the current_hd
      for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
      {
        nd = nt->NodeArray[node_i];

        if ( (nd->extant_leaf == 1) && 
	     (nt->outgroup != nd) &&
	     (hd_from_hybrid[node_i] == current_hd) )
	{
	  nd->impacted_by_parents[0] = 0;
	  picked[node_i] = 1;
          desired_parental_ctr++;

	  last_used_hd = current_hd;

          // DEBUGGING
          //printf("NR DEBUG: node with internal id of %d has been picked for parental subtree\n", node_i);
	}
      }

      current_hd++;
    }

    // reset current as it may or may not have been upped at the last
    // round of the while above
    current_hd = last_used_hd;

    // DEBUGGING
    //printf("NR DEBUG: desired parental ctr %d vs desired parental count %d\n", desired_parental_ctr, desired_parental_count);

    if (desired_parental_ctr < desired_parental_count)
    {
      printf("NR ERROR:  Something went awry in the closest 2x hybrid selection -- program aborted\n");
      exit(0);
    }

    num_to_unpick = 0;

    // check to see if we have too many as that would indicate a
    // tie occurred with the current/last_min_hd and we need to resolve 
    // that situation
    if (desired_parental_ctr > desired_parental_count)
    {
      // DEBUGGING
      //printf("NR DEBUG: we have a situation here ... too many picked....\n");

      num_to_unpick = desired_parental_ctr - desired_parental_count;

      // we will be working our way backwards from the biggest hd
      // which was used for picking originally
      current_hd = last_used_hd;

      num_with_current_hd = 0;
      for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
      {
        if (hd_from_hybrid[node_i] == current_hd)
	{
	  num_with_current_hd++;
	}
      }

      while ( (num_to_unpick > 0) && (current_hd >= 0) )
      {
	num_tries = 0;
	max_num_tries = INT_MAX;
	num_found_this_round = 0;

        // DEBUGGING
        //printf("NR DEBUG: num to unpick is %d and  current hd as %d \n", num_to_unpick, current_hd);

	while ( (num_to_unpick > 0) && 
	        (num_found_this_round < num_with_current_hd) && 
		(num_tries < max_num_tries) )
	{
          which_ctr = 0;
          which_to_unpick = (int) (genrand_real2() * num_with_current_hd);

          //printf("NR DEBUG: inside we have.....num to unpick is %d and which to unpick is %d and with num found this round %d (with num with current hd as %d and current hd %d )\n", num_to_unpick, which_to_unpick, num_found_this_round, num_with_current_hd, current_hd);

	  // find the one to unpick and unpick it if still labeled picked
          for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
          {
            if (hd_from_hybrid[node_i] == current_hd)
	    {
	      if ( (which_to_unpick == which_ctr) &&
	           (picked[node_i] == 1) )
	      {
	        picked[node_i] = 0;
	        nt->NodeArray[node_i]->impacted_by_parents[0] = -1;

                num_to_unpick--;
                desired_parental_ctr--;
		num_found_this_round++;

                // DEBUGGING
                //printf("NR DEBUG: we are unpicking node internal id %d\n", node_i);
	      }
	      which_ctr++;
	    }
	  }
	  num_tries++;
        }

	if (num_to_unpick > 0)
	{
	  current_hd--;

          num_with_current_hd = 0;
          for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
          {
            if (hd_from_hybrid[node_i] == current_hd)
	    {
	      num_with_current_hd++;
	    }
          }
	} // try next hd down
      } // keep looking for ones to unpick
    } // if we have too many picked

    if ( (num_to_unpick > 0) || (current_hd < 0) ||
         (desired_parental_ctr != desired_parental_count) )
    {
      printf("NR ERROR:  Something has gone seriously awry with the 2x algorithm for identifying parental offspring -- program aborted\n");

      // DEBUGGING
      //printf("NR DEBUG: num to unpick %d   current hd %d    ctr %d     count %d\n", num_to_unpick, current_hd, desired_parental_ctr, desired_parental_count);
      exit(0);
    }
  } // we have enough nodes to pick from

  return;

}

/******************************************************************************/
/*                                CheckSubsets                                */
/******************************************************************************/
void CheckSubsets(NETWORK nt)
{
  // WARNING:  remember all this is hard-coded for ONE hybrid

  num_not_impacted = CountExtantNotImpacted(nt);
  num_impacted_by_hybrid = CountExtantImpactedByHybrid(nt);
  num_impacted_by_parents = CountExtantImpactedByParents(nt);

  if (parental_offspring_id_method == 3)
  {
    num_impacted_by_p1 = CountExtantImpactedByParent(nt, -2);
    num_impacted_by_p2 = CountExtantImpactedByParent(nt, -3);
  }
  else
  {
    num_impacted_by_p1 = PARENTAL_AFF_DEFAULT;
    num_impacted_by_p2 = PARENTAL_AFF_DEFAULT;
  }

  if ( (num_not_impacted + num_impacted_by_hybrid + num_impacted_by_parents) !=
       (nt->num_unique_extant_leaves) )
  {
    printf("NR ERROR:  Our subsets (not impacted, impacted by parents, impacted by hybrid) do not add up to match the number of extant leaves -- program aborted\n");
    exit(0);
  }

  // make sure we have at least one taxa not impacted
  if (num_not_impacted < 1)
  {
    printf("NR ERROR:  No extant taxa remain for the overall network construction -- this is not allowed (there must be at least one) -- program aborted\n");
    exit(0);
  }

  // make sure the outgroup taxon is not impacted at all as 
  // it needs to be external to the situation
  if ( (nt->outgroup->impacted_by_parents[0] != -1) ||
       (nt->outgroup->impacted_by_hybrid[0] != -1) )
  {
    printf("NR ERROR:  Outgroup (for the network which is used for the parental subtree and the overall network) cannot be impacted by either the hybrid or its parents -- this is not allowed -- program aborted\n");
    exit(0);
  }

  // make sure the numbers add up if doing extreme custom
  if (parental_offspring_id_method == 3)
  {
    if ( (num_impacted_by_p1 + num_impacted_by_p2) != num_impacted_by_parents)
    {
      printf("NR ERROR:  Number of extant taxa impacted by parents must equal the number of extant taxa affiliated with both parents since this is the extreme custom option -- this has been violated -- program aborted\n");
      exit(0);
    }
  }
 
}

/******************************************************************************/
/*                           ReconParentalTopology                            */
/******************************************************************************/
void ReconParentalTopology(NETWORK nt, int psub)
{
  int subtree_id;
  int node_i;
  NODE og;
  FILE *pipe_ptr;

  // WARNING:  remember this is all for one hybrid subtree and its parents
  //           subtree refers to hybrid number, while psub refers to the
  //           parental subtree (0 or 1) 
  subtree_id = 0;

  // see if we will have a trivial parental subtree -- meaning no taxa
  // impacted by parents -- then there is nothing to do practically
  // (whereas if there is at least one then that plus the outgroup
  // can get reconstructed using parsimony -- a minimum of three is
  // needed for nj reconstruction which is controlled by the threshold
  // which makes sure the minimum is met)
  if ( (parental_offspring_id_method != 3) && 
       (num_impacted_by_parents == 0) )
  {
    CreateTrivialParentalSubtree(nt);
  }
  else if ( (parental_offspring_id_method == 3) && 
            (psub == 0) && 
	    (num_impacted_by_p1 == 0) )
  {
    CreateTrivialParentalSubtree(nt);
  }
  else if ( (parental_offspring_id_method == 3) && 
            (psub == 1) && 
	    (num_impacted_by_p2 == 0) )
  {
    CreateTrivialParentalSubtree(nt);
  }
  else
  {
    // first find the outgroup for the network and save it so we
    // can pass it to the write function
    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      if (nt->NodeArray[node_i] == nt->outgroup)
      {
        og = nt->NodeArray[node_i];
      }
    }

    if (og == NULL)
    {
      printf("NR ERROR:  Could not find outgroup for network (to be used with parental subtree reconstruction) -- program aborted\n");
      exit(0);
    }

    // we must reset the writing markers and set the ones that 
    // should be set for this parental topology work
    ResetWriteMarkers(nt);

    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      if (parental_offspring_id_method == 3)
      {
        if ( (nt->NodeArray[node_i]->impacted_by_parents[0] != -1) &&
	     (nt->NodeArray[node_i]->parental_aff == (PARENTAL_AFF_DEFAULT - (psub + 1))) )
        {
          nt->NodeArray[node_i]->marked_to_write = 1;
        }
      }
      else // all others just look at whether impacted by parents or not
      {
        if (nt->NodeArray[node_i]->impacted_by_parents[0] != -1)
        {
          nt->NodeArray[node_i]->marked_to_write = 1;
        }
      }
    }

    // now we can write the input file
    if (parental_offspring_id_method != 3)
    {
      WriteInputFile(nt, num_impacted_by_parents, og, subtree_id);
    }
    else
    {
      if (psub == 0)
      {
        WriteInputFile(nt, num_impacted_by_p1, og, subtree_id);
      }
      else // second parent subtree
      {
        WriteInputFile(nt, num_impacted_by_p2, og, subtree_id);
      }
    }

    // now run the appropriate reconstruction
    if (recon_type == 0) // PHYLIP
    {
      if ( ((parental_offspring_id_method != 3) && 
            (num_impacted_by_parents <= mp_phylip_thresh)) ||
	   ((parental_offspring_id_method == 3) &&
	    (psub == 0) &&
	    (num_impacted_by_p1 <= mp_phylip_thresh)) ||
	   ((parental_offspring_id_method == 3) &&
	    (psub == 1) &&
	    (num_impacted_by_p2 <= mp_phylip_thresh)) )
      {
        ExecuteDnaPenny(nt, subtree_id);

        if ((pipe_ptr = popen("mv outfile outfile_dna_penny \n mv outtree outtree_dna_penny\n", "r")) != NULL)
        {
          (void) pclose(pipe_ptr);
        }
        else
        {
          printf("NR ERROR:  Unexpected issue when moving outtree or outfile after running PHYLIP's dnapenny to create the parental subtree -- program aborted\n");
	  exit(0);
        }

        ExecuteConsense(nt, subtree_id); // doing Majority-Rule extended
      }
      else
      {
        ExecuteDnaDist(nt);

        if ((pipe_ptr = popen("mv outfile outfile_dna_dist \n", "r")) != NULL)
        {
          (void) pclose(pipe_ptr);
        }
        else
        {
          printf("NR ERROR:  Unexpected issue when moving outfile after running PHYLIP's dnadist during parental subtree work -- program aborted\n");
	  exit(0);
        }

        ExecuteNeighbor(nt);
      }
    }
    else if (recon_type == 1) // RANDIP
    {
      ExecuteRandIP(nt, subtree_id);
    }
    else
    {
      printf("NR ERROR:  Invalid reconstruction algorithm chosen for parental subtree -- program aborted\n");
      exit(0);
    }

    // now rename final output files
    if ((pipe_ptr = popen("mv outfile outfile_parental_subtree \n mv outtree outtree_parental_subtree\n", "r")) != NULL)
    {
      (void) pclose(pipe_ptr);
    }
    else
    {
      printf("NR ERROR:  Unexpected issue when moving outtree or outfile after running PHYLIP's consense/neighbor (or RANDIP) to create parental subtree -- program aborted\n");
      exit(0);
    }
  }

  return;
}

/******************************************************************************/
/*                         CreateTrivialParentalSubtree                       */
/******************************************************************************/
void CreateTrivialParentalSubtree(NETWORK nt)
{
  int na_index;

  // WARNING:  remember all this is customized for ONE hybrid

  // first make the root
  na_index = nt->node_id_ctr++;
  nt->parental_subtree_root = NewNode(na_index);
  nt->parental_subtree_root->temporary = 1;
  nt->NodeArray[na_index] = nt->parental_subtree_root;
  nt->num_unique_nodes++;

  // to be consistent with the non-trivial reconstruction, the root
  // is temporarily set to the subtree root (as if we had done
  // newick parsing)
  nt->root = nt->parental_subtree_root;

  // nothing else to do
  return;

}

/******************************************************************************/
/*                          ReadParentalSubtreeTopology                       */
/******************************************************************************/
void ReadParentalSubtreeTopology(NETWORK nt)
{
  // first we need to open the outtree file
  fp_temp_in = fopen("outtree_parental_subtree", "r");

  if (fp_temp_in == NULL)
  {
    printf("NR ERROR:  Cannot open the temp input file for reading parental subtree results (outtree_parental_subtree) -- program aborted\n");
    exit(0);
  }

  // now we can do the parsing
  ParseNewick(nt);

  // finally close the file
  fclose(fp_temp_in);

  return;

}

/******************************************************************************/
/*                        UnhookParentalSubtreeOutgroup                       */
/******************************************************************************/
void UnhookParentalSubtreeOutgroup(NETWORK nt)
{
  NODE og;

  og = nt->outgroup;

  // first deal with the direction from parent to children
  if (nt->parental_subtree_root->first_child_node == og)
  {
    if (og->next_sib_node == NULL)
    {
      printf("NR ERROR:  Something went awry with the parental subtree topology as the outgroup is the only child of the root -- program aborted\n");
      exit(0);
    }
    else
    {
      nt->parental_subtree_root->first_child_node = og->next_sib_node; 
      og->next_sib_node->prev_sib_node = NULL;
    }
  }
  else // og is a sibling
  {
    if (og->next_sib_node != NULL)
    {
      og->next_sib_node->prev_sib_node = og->prev_sib_node;
      og->prev_sib_node->next_sib_node = og->next_sib_node;
    }
    else
    {
      og->prev_sib_node->next_sib_node = NULL;
    }
  }

  // now reset outgroup values
  og->parent1_node = NULL;
  og->next_sib_node = NULL;
  og->prev_sib_node = NULL;


  return;
}

/******************************************************************************/
/*                           AssignParentalLineages                           */
/******************************************************************************/
void AssignParentalLineages(NETWORK nt)
{
  NODE nd;
  int lin_ctr, i, node_i;
  double rand_val, avgp1, avgp2;

  // first see if we need to move the root down one level now that
  // the outgroup has been removed -- we want the root to be where
  // the parental lineages will originate

  if ( (num_impacted_by_parents > 1) &&
       (nt->parental_subtree_root->first_child_node->next_sib_node == NULL) )
  {
    nt->parental_subtree_root = nt->parental_subtree_root->first_child_node;
    nt->root = nt->parental_subtree_root;
    nt->parental_subtree_root->temporary = 1;

    // DEBUGGING
    //printf("NR DEBUG: we had to move down one notch\n");
  }

  // we should be okay now...
  if ( (num_impacted_by_parents > 1) &&
       (nt->parental_subtree_root->first_child_node->next_sib_node == NULL) )
  {
    printf("NR ERROR:  There was a problem with the parental subtree reconstruction -- the SUB-root only has one child -- program aborted\n");
    exit(0);
  }

  ResetMarkersAll(nt);

  // now mark the parental lineages for reference and count them
  // but start at 1 so we don't get messed with our counter stuff!
  lin_ctr = 1;

  nd = nt->parental_subtree_root->first_child_node;
  nd->marker = -1;
  lin_ctr++;

  nd = nd->next_sib_node;
  while (nd != NULL)
  {
    nd->marker = -1;
    lin_ctr++;
    nd = nd->next_sib_node;
  }

  // DEBUGGING
  //printf("NR DEBUG: num of parental lineages in the parental subtree %d\n", lin_ctr-1);

  // for each parental lineage, label the descendant taxa
  // calculate that group's extant taxa avg hamming distance
  // from each parent node (as labeled with half seqs in
  // hybrid subtree structure) and finally label the 
  // parental lineage here with whether that clade will
  // associate with parent1 or parent2
  for (i = 1; i <= lin_ctr; i++)
  {
    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      if (nt->NodeArray[node_i]->marker == -1)
      {
        SetMarkersTopo(nt->NodeArray[node_i], i);

	avgp1 = CalcAvgHamFromParent(nt, i, nt->hybrid_subtree_root->parent1_node);
	avgp2 = CalcAvgHamFromParent(nt, i, nt->hybrid_subtree_root->parent2_node);

        // DEBUGGING
        //printf("NR DEBUG: avgp1 %f avgp2 %f\n", avgp1, avgp2);

        ResetMarkersTopo(nt->NodeArray[node_i]);

        if (avgp1 < avgp2)
	  nt->NodeArray[node_i]->marker = 1;
	else if (avgp2 < avgp1)
	  nt->NodeArray[node_i]->marker = 2;
	else // tie
	{
	  rand_val = genrand_real1();

          // DEBUGGING
          //printf("NR DEBUG: rand val for tie is %f\n", rand_val);

	  if (rand_val < .5)
	    nt->NodeArray[node_i]->marker = 1;
	  else
	    nt->NodeArray[node_i]->marker = 2;
	}
      }
    }
  }

  return;

}

/******************************************************************************/
/*                        AssignParentalLineagesExtreme                       */
/******************************************************************************/
void AssignParentalLineagesExtreme(NETWORK nt, int psub)
{
  NODE new_root;
  NODE old_root;
  int na_index;

  // first see if the first child has a sibling -- if it does not,
  // we are in good shape as this will be the subtree's first
  // child and get attached correctly to the specified parent,
  // what we care about and must extend up one level so we get a single
  // extension for connecting to the parent -- this routine is
  // called only if there is at least one, so can assume that part

  
  if (nt->parental_subtree_root->first_child_node->next_sib_node != NULL)
  {
    na_index = nt->node_id_ctr++;
    new_root = NewNode(na_index);

    nt->num_unique_nodes = nt->num_unique_nodes + 1;
    nt->NodeArray[na_index] = new_root;

    old_root = nt->parental_subtree_root;
    old_root->parent1_node = new_root;

    new_root->first_child_node = old_root;

    nt->root = new_root;
    nt->parental_subtree_root = new_root;
    nt->parental_subtree_root->temporary = 1;
  }

  // double check to make sure all is ok
  if  (nt->parental_subtree_root->first_child_node->next_sib_node != NULL)
  {
    printf("NR ERROR:  There was a problem with the parental subtree reconstruction for the extreme case -- the root has more than one child -- program aborted\n");
    exit(0);
  }

  ResetMarkersAll(nt);

  // since this is the extreme case, we know exactly which parent this
  // subtree will get attached to, so simply set the marker for the 
  // root's first (and only) child correctly

  nt->parental_subtree_root->first_child_node->marker = psub + 1;

  return;

}


/******************************************************************************/
/*                          CalcAvgHamFromParent                              */
/******************************************************************************/
double CalcAvgHamFromParent(NETWORK nt, int marker_value, NODE parent_nd)
{
  int denom, node_i, sum, hd;
  double avg_ham;
  NODE nd;

  sum = 0;
  denom = 0;

  // go through, identify, count, and calc individual hd and tally along the way
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    nd = nt->NodeArray[node_i];

    if ( (nd->extant_leaf == 1) && (nd->marker == marker_value) )
    {
      hd = HammingDistanceTotal(nd, parent_nd);  
      sum = sum + hd;
      denom++;

      // DEBUGGING
      //printf("NR DEBUG:just did hd from pid %d and hd is %d and sum is %d and denom is %d (marker value is %d)\n", parent_nd->internal_node_id, hd, sum, denom, marker_value);

    }
  }

  avg_ham = (double) ( ((double) sum)/((double) denom) );

  return avg_ham;
}

/******************************************************************************/
/*                           ConnectSubtrees                                  */
/******************************************************************************/
void ConnectSubtrees(NETWORK nt)
{
  NODE p1;
  NODE p2;
  NODE nd;
  NODE cn;

  int node_i;

  // DEBUGGING
  //printf("NR DEBUG: starting connect subtrees\n");

  // here we look for the clades labeled with which parent they will
  // associate -- then we have to hook these up as children of the
  // parent nodes already existing in the hybrid subtree structure

  p1 = nt->hybrid_subtree_root->parent1_node;
  p2 = nt->hybrid_subtree_root->parent2_node;
  
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    nd = nt->NodeArray[node_i];

    // DEBUGGING
    //printf("NR DEBUG: inside connect subtrees and for node %d its marker is %d\n", node_i, nd->marker);

    // the marker tells us which parent the lineage will belong to
    if (nd->marker == 1)
    {
      // DEBUGGING
      //printf("NR DEBUG:  have found node %d to be p1's kid\n", nd->internal_node_id);

      // we know that p1's first child is the reticulate node, 
      // but we must determine if this node will be the first p1 sibling
      // if so need to set it as such, if not need to find
      // the end of the list and set it there
      cn = nt->hybrid_subtree_root;

      if (cn->next_p1_sib == NULL)
      {
        cn->next_p1_sib = nd;
	nd->prev_sib_node = cn;
	nd->next_sib_node = NULL;
	nd->parent1_node = nt->hybrid_subtree_root->parent1_node;
      }
      else
      {
        cn = cn->next_p1_sib;

	while (cn->next_sib_node != NULL)
	{
	  cn = cn->next_sib_node;
	}

	// now at end of line
	cn->next_sib_node = nd;
	nd->prev_sib_node = cn;
	nd->next_sib_node = NULL;
	nd->parent1_node = nt->hybrid_subtree_root->parent1_node;
      }
    } // will be a p1 lineage

    if (nd->marker == 2)
    {
      // DEBUGGING
      //printf("NR DEBUG:  have found node %d to be p2's kid\n", nd->internal_node_id);
      // we know that p2's first child is the reticulate node, 
      // but we must determine if node will be the first p2 sibling
      // if so need to set it as such, if not need to find
      // the end of the list and set it there
      cn = nt->hybrid_subtree_root;

      if (cn->next_p2_sib == NULL)
      {
        cn->next_p2_sib = nd;
	nd->prev_sib_node = cn;
	nd->next_sib_node = NULL;
	nd->parent1_node = nt->hybrid_subtree_root->parent2_node;
      }
      else
      {
        cn = cn->next_p2_sib;

	while (cn->next_sib_node != NULL)
	{
	  cn = cn->next_sib_node;
	}

	// now at end of line
	cn->next_sib_node = nd;
	nd->prev_sib_node = cn;
	nd->next_sib_node = NULL;
	nd->parent1_node = nt->hybrid_subtree_root->parent2_node;
      }
    } // will be a p2 lineage
  } // next node

  return;
}
