/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<float.h>
#include<limits.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"twister.h"
#include"phylipinter.h"


/******************************************************************************/
/*                            WriteInputFile                                  */
/******************************************************************************/
void WriteInputFile(NETWORK nt, int ctr, NODE og, int subtree_id)
{
  int num_of_seqs, num_secs_per_line, total_num_bps;
  int node_i, line_i, bps_per_section;
  int num_of_lines;
  char phylip_input_file[14] = "phylip_infile";

  // first open the file
  fp_temp_out = fopen(phylip_input_file,"w");
  if (fp_temp_out == NULL)
  {
    printf("NR ERROR:  Cannot open the temp output file for input into PHYLIP -- program aborted\n");
    exit(0);
  }

  // some values we need for writing -- assume 50 base pairs per line
  num_of_seqs = num_hom_chromos * num_strands_per_hom_chrom;
  num_secs_per_line = 5;
  bps_per_section = 5;
  total_num_bps = num_of_seqs * sequence_length;
  num_of_lines = (int) (ceil) ( ((double)total_num_bps) / ((double) (num_secs_per_line*bps_per_section)) );

  // first write the number of taxa and the number of sites for each
  // note that the outgroup is always one of the taxa not impacted
  // but needs to be listed in the input file
  fprintf(fp_temp_out, "   %d    %d\n", ctr + 1, total_num_bps); 

  // we need to list the outgroup for this subtree as the first species
  // because later when writing option files we always assume it is first
  WriteFirstLine(og, num_secs_per_line, bps_per_section, total_num_bps);

  // now for each extant taxa that will be part of this file we list it
  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    if (nt->NodeArray[node_i]->marked_to_write == 1)
    {
      WriteFirstLine(nt->NodeArray[node_i], num_secs_per_line, bps_per_section, total_num_bps);

    }
  } // next node in NodeArray

  // put a line with only a carriage return after the first set of lines
  fprintf(fp_temp_out,"\n");       

  // once the first line is written for the outgroup and all the
  // extant taxa to be listed, need to do the remaining lines which 
  // just contain the remaining DNA sequence info
  
  for (line_i = 1; line_i < num_of_lines; line_i++)
  {
    // first do the outgroup
    WriteSubsequentLines(og, num_secs_per_line, bps_per_section, total_num_bps, line_i);

    // then do the others
    for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
    {
      if (nt->NodeArray[node_i]->marked_to_write == 1)
      {
        WriteSubsequentLines(nt->NodeArray[node_i], num_secs_per_line, bps_per_section, total_num_bps, line_i);

      }
    }

    // at the end of this node grouping put a carriage return
    fprintf(fp_temp_out,"\n");       
  }

  // finally close the file
  fclose(fp_temp_out);

  return;
}

/******************************************************************************/
/*                          WriteFirstLine                                    */
/******************************************************************************/
void WriteFirstLine(NODE nd, int num_secs_per_line,
                    int bps_per_section, int total_num_bps)
{
  int i, j, bp_i, hchromo_i, strand_i, num_blanks;
  char id_str[11];
  char temp_all_bps[MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_SEQ_LENGTH];


  strcpy(temp_all_bps, "\0"); 

  // to make things easier concatenate all the sequence info together
  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      strcat(temp_all_bps, nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence); 
    }
  }

  // NOTE: we use internal ids for all PHYLIP processing
  //       this is because the internal nodes will not
  //       have unique external ids
  sprintf(id_str, "%d", nd->internal_node_id);

  // now we need to add the correct number of blanks to the name
  num_blanks = 10 - strlen(id_str);

  for (i = strlen(id_str); i < 10; i++)
  {
    strcpy(&(id_str[i]), " ");
  }

  // now we can finally write the name
  fprintf(fp_temp_out, "%s", &id_str[0]);

  bp_i = 0;

  // now do only the first line of the dna seq(s)
  for (i = 0; i < num_secs_per_line; i++)
  {
    for (j = 0; j < bps_per_section; j++)
    {
      if (bp_i < total_num_bps)
      {
        fprintf(fp_temp_out,"%.*s", 1, &temp_all_bps[bp_i]);       
        bp_i++;
      }
    }

    // put a space after so many characters
    fprintf(fp_temp_out," ");       
  }

  // at the end of the line put a carriage return
  fprintf(fp_temp_out,"\n");       

  return;
}

/******************************************************************************/
/*                          WriteSubsequentLines                              */
/******************************************************************************/
void WriteSubsequentLines(NODE nd, int num_secs_per_line, int bps_per_section, 
                          int total_num_bps, int line_i)
{
  int i, j, bp_i, hchromo_i, strand_i;
  char temp_all_bps[MAX_NUM_HOM_CHROMOS*MAX_NUM_STRANDS_PER_HC*MAX_SEQ_LENGTH];

  strcpy(temp_all_bps, "\0"); 

  // to make things easier concatenate all the sequence info together
  for (hchromo_i = 0; hchromo_i < num_hom_chromos; hchromo_i++)
  {
    for (strand_i = 0; strand_i < num_strands_per_hom_chrom; strand_i++)
    {
      strcat(temp_all_bps, nd->hom_chromos[hchromo_i]->strands[strand_i]->sequence); 
    }
  }

  bp_i = line_i * num_secs_per_line * bps_per_section;

  for (i = 0; i < num_secs_per_line; i++)
  {
    for (j = 0; j < bps_per_section; j++)
    {
      if (bp_i < total_num_bps)
      {
        fprintf(fp_temp_out,"%.*s", 1, &temp_all_bps[bp_i]);       
        bp_i++;
      }
    }
    // put a space after so many characters
    fprintf(fp_temp_out," ");       
  }

  // at the end of the line put a carriage return
  fprintf(fp_temp_out,"\n");       

  return;
}

/******************************************************************************/
/*                          ExecuteDnaPenny                                   */
/******************************************************************************/
void ExecuteDnaPenny(NETWORK nt, int subtree_id)
{
  FILE *pipe_ptr;

  // first write the options file which includes the outgroup
  WriteDnaPennyOptFile(nt, subtree_id);

  // now execute dnapenny
  if ((pipe_ptr = popen("phylip dnapenny < dna_penny_opt > rc_dnapenny\n", "r")) != NULL)
  {
    // close the pipe
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NR ERROR:  Unexpected issue when running PHYLIP's dnapenny -- program aborted\n");
    exit(0);
  }

  return;
}

/******************************************************************************/
/*                          WriteDnaPennyOptFile                              */
/******************************************************************************/
void WriteDnaPennyOptFile(NETWORK nt, int subtree_id)
{
  int subtree_outgroup_index;

  // when writing the input file we always put the outgroup first
  subtree_outgroup_index = 1;

  // first open the file
  fp_temp_out = fopen("dna_penny_opt","w");

  if (fp_temp_out == NULL)
  {
    printf("NR ERROR:  Cannot open the temp output file for DNA Penny options file -- NR aborted\n");
    exit(0);
  }

  fprintf(fp_temp_out, "phylip_infile\n"); // give the input file name
  fprintf(fp_temp_out, "s\n");  // reconsiders order of species 
  fprintf(fp_temp_out, "o\n");  // outgroup
  fprintf(fp_temp_out, "%d\n", subtree_outgroup_index);  // outgroup index
  fprintf(fp_temp_out, "4\n");  // prints out steps in each site
  fprintf(fp_temp_out, "5\n");  // prints out seqs at all nodes
  fprintf(fp_temp_out, "Y\n");  // go ahead and execute

  // finally close the file
  fclose(fp_temp_out);

  return;
}

/******************************************************************************/
/*                             ExecuteDnaDist                                 */
/******************************************************************************/
void ExecuteDnaDist(NETWORK nt)
{
  FILE *pipe_ptr;
  char command_line[500];

  // first write the option file
  WriteDnaDistOptFile(nt);

  // now execute
  //if ((pipe_ptr = popen("phylip dnadist < dna_dist_opt > rc_dna_dist\n", "r")) != NULL)

  if (dna_dist_model != 4)
  {
    strcpy(command_line, "phylip dnadist < dna_dist_opt > rc_dna_dist\n");
  }
  else // custom option #4
  {
    strcpy(command_line, "phylip dnadist < ");
    strcat(command_line, dna_dist_custom_file_name);
    strcat(command_line, " > rc_dna_dist\n");
  }

  // DEBUGGING
  //printf("command line for dna dist is:%s\n", command_line);

  if ((pipe_ptr = popen(command_line, "r")) != NULL)
  {
    // close the pipe
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NR ERROR:  Unexpected issue when running PHYLIP's dnadist -- program aborted\n");
    exit(0);
  }

  return;
}

/******************************************************************************/
/*                              WriteDnaDistOptFile                           */
/******************************************************************************/
void WriteDnaDistOptFile(NETWORK nt)
{
  // the writing of this option file is more complicated than most because
  // an evolutionary model should be specified in it so PHYLIP knows how
  // to operate -- and we let the user specify one of four "generic" options
  // OR they can give their own.....

  // if not specifed -- go with default of jukes cantor
  if (dna_dist_model_yn == 0)
  {
    dna_dist_model = 0;
  }
 
  if (dna_dist_model != 4) // if don't have custom to deal with...
  {
    // first open the file
    fp_temp_out = fopen("dna_dist_opt", "w");

    if (fp_temp_out == NULL)
    {
      printf("NR ERROR:  Cannot open temp output file for DNA Dist options file -- NR aborted\n");
      exit(0);
    }

    // now do the correct writing depending on model requested

    if (dna_dist_model == 0) // Jukes-Cantor
    {
      fprintf(fp_temp_out, "phylip_infile\n"); // give the input file name
      fprintf(fp_temp_out, "d\n"); // switch -- need both d's here to get
      fprintf(fp_temp_out, "d\n"); // phylip to switch to jukes-cantor
      fprintf(fp_temp_out, "Y\n"); // go ahead and execute
    }
    else if (dna_dist_model == 1) // f84
    {
      fprintf(fp_temp_out, "phylip_infile\n"); // give the input file name
      fprintf(fp_temp_out, "Y\n"); // no switching needed, go ahead and execute
    }
    else if (dna_dist_model == 2) // k2p
    {
      fprintf(fp_temp_out, "phylip_infile\n"); // give the input file name
      fprintf(fp_temp_out, "d\n"); // switch -- need one d here to get k2p
      fprintf(fp_temp_out, "Y\n"); // go ahead and execute
    }
    else if (dna_dist_model == 3) // logdet
    {
      fprintf(fp_temp_out, "phylip_infile\n"); // give the input file name
      fprintf(fp_temp_out, "d\n"); // switch -- need all three d's here to get
      fprintf(fp_temp_out, "d\n"); // log det model
      fprintf(fp_temp_out, "d\n"); // 
      fprintf(fp_temp_out, "Y\n"); // go ahead and execute
    }
    // finally close the file
    fclose(fp_temp_out);
  }
  else if (dna_dist_model == 4) // custom option -- do nothing for now
  {
    ;
  }
  else
  {
    printf("NR ERROR:  problem with the logistics of writing dna dist option file -- invalid model speficied -- program aborted\n");
    exit(0);
  }

  return;
}

/******************************************************************************/
/*                             ExecuteNeighbor                                */
/******************************************************************************/
void ExecuteNeighbor(NETWORK nt)
{
  FILE *pipe_ptr;
  // first write the option file
  WriteNeighborOptFile(nt);

  // now execute
  if ((pipe_ptr = popen("phylip neighbor < neighbor_opt > rc_neighbor\n", "r")) != NULL)
  {
    // close the pipe
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NR ERROR:  Unexpected issue when running PHYLIP's neighbor -- program aborted\n");
    exit(0);
  }

  return;
}

/******************************************************************************/
/*                              WriteNeighborOptFile                          */
/******************************************************************************/
void WriteNeighborOptFile(NETWORK nt)
{
  int og_index;
  int rand_val, rand_problem, rand_try_ctr;

  // first open the file
  fp_temp_out = fopen("neighbor_opt", "w");

  if (fp_temp_out == NULL)
  {
    printf("NR ERROR:  Cannot open temp output file for Neighbor options file -- NR aborted\n");
    exit(0);
  }

  // when writing the input file we always put the outgroup first
  og_index = 1;

  // we need a random number seed for the jumble option -- must be
  // between 0 and 32767 and must be of the form 4n + 1 -- so
  // do the work here....
  rand_try_ctr = 0;
  rand_problem = 1;

  while ( (rand_problem == 1) && (rand_try_ctr < 5000000) )
  {
    rand_val = (int) genrand_int32();
    rand_try_ctr++;

    if ( ((rand_val % 4) != 1) || (rand_val < 0) || (rand_val > 32767) )
      rand_problem = 1;
    else
      rand_problem = 0;
  }

  if ( (rand_try_ctr >= 5000000) && (rand_problem == 1) )
  {
    printf("NR ERROR:  Could not generate a suitable random seed for PHYLIP's neighbor program (jumble option) after 5,000,000 tries -- program aborted\n");
    exit(0);
  }

  fprintf(fp_temp_out, "outfile_dna_dist\n"); // give the input file name

  fprintf(fp_temp_out, "o\n");            // outgroup
  fprintf(fp_temp_out, "%d\n", og_index); // outgroup
  fprintf(fp_temp_out, "j\n");            // jumble the input
  fprintf(fp_temp_out, "%d\n", rand_val); // give seed for jumblet
  fprintf(fp_temp_out, "Y\n"); // go ahead and execute

  // finally close the file
  fclose(fp_temp_out);
}

/******************************************************************************/
/*                          ExecuteConsense                                   */
/******************************************************************************/
void ExecuteConsense(NETWORK nt, int subtree_id)
{
  FILE *pipe_ptr;

  // first write the options file
  WriteConsenseOptFile(nt, subtree_id);

  // now execute consense
  if ((pipe_ptr = popen("phylip consense < dna_consense_opt > rc_consense\n", "r")) != NULL)
  {
    // close the pipe
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NR ERROR:  Unexpected issue when running PHYLIP's consense -- program aborted\n");
    exit(0);
  }
  return;
}


/******************************************************************************/
/*                          WriteConsenseOptFile                              */
/******************************************************************************/
void WriteConsenseOptFile(NETWORK nt, int subtree_id)
{
  // first open the file
  fp_temp_out = fopen("dna_consense_opt","w");

  if (fp_temp_out == NULL)
  {
    printf("NR ERROR:  Cannot open the temp output file for Consense options file\n");
    exit(0);
  }

  // doing extended majority rule consensus which is the default
  fprintf(fp_temp_out, "outtree_dna_penny\n"); // give the input file name

  fprintf(fp_temp_out, "r\n");  // so that the trees are read in as rooted
  fprintf(fp_temp_out, "Y\n");  // go ahead and execute

  // finally close the file
  fclose(fp_temp_out);

  return;
}

