/*****************************************************************************/
/*                                                                           */
/*  NetReconstruct kappa version                                             */
/*                                                                           */
/*  Software to perform phylogenetic network reconstructions when            */
/*  diploid hybrids are part of the evolutionary history.                    */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include"globals.h"
#include"node.h"
#include"network.h"
#include"phylipinter.h"
#include"randipinter.h"
#include"output.h"
#include"hybsub.h"
#include"twister.h"
#include"rem.h"


/******************************************************************************/
/*                               ReconRemainder                               */
/******************************************************************************/
void ReconRemainder(NETWORK nt)
{
  // now that we have our subtree + extended structure +
  // parental subtree, it is time to deal with the remainder
  ReconRemainderTopology(nt);

  // now we can read the results
  ReadRemTopology(nt);

  // if the outgroup adjust flag is active, check to see if
  // the extra work is needed and do it...
  if (outgroup_adjustment == 1)
  {
    OutgroupAdjust(nt);
  }

  return;
} 

/******************************************************************************/
/*                        ReconRemainderTopology                              */
/******************************************************************************/
void ReconRemainderTopology(NETWORK nt)
{
  FILE *pipe_ptr;
  NODE og;
  int node_i, subtree_id;

  // no longer dealing with subtrees here -- placeholder
  subtree_id = -1;

  // first we need to label the not impacted and parental hybrid nodes as 
  // marked so that we will know which nodes to write for this the next
  // step of the PHYLIP work (outgroup is handled separately)
  ResetWriteMarkers(nt);

  for (node_i = 0; node_i < nt->num_unique_nodes; node_i++)
  {
    // looking for extant leaves that were not involved/impacted
    // by the hybrid subtree or parental subtree
    if ( (nt->NodeArray[node_i]->impacted_by_hybrid[0] == -1) &&
         (nt->NodeArray[node_i]->impacted_by_parents[0] == -1) &&
         (nt->NodeArray[node_i]->extant_leaf == 1) &&
	 (nt->NodeArray[node_i] != nt->outgroup) )
    {
      nt->NodeArray[node_i]->marked_to_write = 1;
    }
  }

  // now do the parent marking as they are not extant leaves, but
  // will act as such for PHYLIP's neighbor work
  nt->hybrid_subtree_root->parent1_node->marked_to_write = 1;
  nt->hybrid_subtree_root->parent2_node->marked_to_write = 1;

  num_not_impacted = CountExtantNotImpacted(nt);

  num_not_impacted = num_not_impacted + 2; // count the parental nodes

  num_not_impacted = num_not_impacted - 1; // subtract one because write input
                                           // will add one for the outgroup
					   // and we have already counted it
					   // in this case

  og = nt->outgroup;

  // then we need to write the PHYLIP file for the dnadist program
  WriteInputFile(nt, num_not_impacted, og, subtree_id);

  // DEBUGGING
  if ((pipe_ptr = popen("cp phylip_infile phylip_infile_now \n", "r")) != NULL)
  {
    (void) pclose(pipe_ptr);
  }

  // DEBUGGING
  //printf("NR DEBUG: num not impacted is %d\n", num_not_impacted);

  // now run the appropriate reconstruction
  if (recon_type == 0) // PHYLIP
  {
    if (num_not_impacted <= mp_phylip_thresh)
    {
      ExecuteDnaPenny(nt, subtree_id);

      if ((pipe_ptr = popen("mv outfile outfile_dna_penny \n mv outtree outtree_dna_penny\n", "r")) != NULL)
      {
        (void) pclose(pipe_ptr);
      }
      else
      {
        printf("NR ERROR:  Unexpected issue when moving outtree or outfile after running PHYLIP's dnapenny to create the remaining structure -- program aborted\n");
        exit(0);
      }

      ExecuteConsense(nt, subtree_id);
    }
    else
    {
      ExecuteDnaDist(nt);

      if ((pipe_ptr = popen("mv outfile outfile_dna_dist \n", "r")) != NULL)
      {
        (void) pclose(pipe_ptr);
      }
      else
      {
        printf("NR ERROR:  Unexpected issue when moving outfile after running PHYLIP's dnadist to create the remaining structure -- program aborted\n");
      exit(0);
      }

      ExecuteNeighbor(nt);
    }
  }
  else if (recon_type == 1) // RANDIP
  {
    ExecuteRandIP(nt, subtree_id);
  }
  else
  {
    printf("NR ERROR:  Invalid reconstruction algorithm chosen for remaining subtree -- program aborted\n");
    exit(0);
  }

  // now rename final output files
  if ((pipe_ptr = popen("mv outfile outfile_remaining \n mv outtree outtree_remaining", "r")) != NULL)
  {
    // close the pipe
    (void) pclose(pipe_ptr);
  }
  else
  {
    printf("NR ERROR:  Unexpected issue when moving outfile after running PHYLIP's consense/neighbor (or RANDIP) for remaining extant -- program aborted\n");
    exit(0);
  }

  return;

}

/******************************************************************************/
/*                               ReadRemTopology                              */
/******************************************************************************/
void ReadRemTopology(NETWORK nt)
{
  // first we need to open the outtree file
  fp_temp_in = fopen("outtree_remaining", "r");

  if (fp_temp_in == NULL)
  {
    printf("NR ERROR:  Cannot open the temp input file for reading remaining subtree results (outtree_remaining) -- program aborted\n");
    exit(0);
  }

  // now we can do the parsing
  ParseNewick(nt);

  // finally close the file
  fclose(fp_temp_in);

  return;

}

/******************************************************************************/
/*                            Outgroup Adjust                                 */
/******************************************************************************/
void OutgroupAdjust(NETWORK nt)
{
  NODE og;
  NODE new_root;
  NODE old_root;
  int na_index;

  // we know the outgroup is one of the roots kids, but we only want to 
  // adjust if the root has three or more kids, indicating the keeping of
  // the outgroup at that level is an unexpected polytomy due to probably
  // neighbor joining...root has a minimum of two kids by definition of 
  // our topology structure
  
  if (nt->root->first_child_node->next_sib_node->next_sib_node != NULL)
  {
    og = nt->outgroup;

    // DEBUGGING
    //printf("NR DEBUG:  about to do outgroup adjust -- og internal id is %d (ext is %d) and root's internal id is %d (ext is %d)\n", og->internal_node_id, og->external_node_id, nt->root->internal_node_id, nt->root->external_node_id);
    
    //printf("NR DEBUG: first child internal id is %d (ext is %d) and its sib internal id is %d (ext %d) and its sib internal id is %d (ext %d)\n", nt->root->first_child_node->internal_node_id, nt->root->first_child_node->external_node_id, nt->root->first_child_node->next_sib_node->internal_node_id, nt->root->first_child_node->next_sib_node->external_node_id, nt->root->first_child_node->next_sib_node->next_sib_node->internal_node_id, nt->root->first_child_node->next_sib_node->next_sib_node->external_node_id);
    // END DEBUGGING

    // first remove the outgroup
    if (nt->root->first_child_node == og)
    {
      if (og->next_sib_node == NULL)
      {
        printf("NR ERROR:  Something got messed up with outgroup adjustment for the remaining tree -- program aborted\n");
	exit(0);
      }
      else
      {
        nt->root->first_child_node = og->next_sib_node;
	og->next_sib_node->prev_sib_node = NULL;
      }
    }
    else // og is not the first kid
    {
      if (og->next_sib_node != NULL)
      {
        og->next_sib_node->prev_sib_node = og->prev_sib_node;
	og->prev_sib_node->next_sib_node = og->next_sib_node;
      }
      else
      {
        og->prev_sib_node->next_sib_node = NULL;
      }
    }

    // be good as we go along here....
    og->parent1_node = NULL;
    og->next_sib_node = NULL;
    og->prev_sib_node = NULL;


    // now make new node to be overall root
    na_index = nt->node_id_ctr++;
    new_root = NewNode(na_index);

    nt->num_unique_nodes = nt->num_unique_nodes + 1;

    old_root = nt->root;
    old_root->parent1_node = new_root;

    // hook up old root and outgroup as children of new root
    new_root->first_child_node = old_root;
    new_root->first_child_node->next_sib_node = og;
    og->prev_sib_node = new_root->first_child_node;
    og->parent1_node = new_root;

    // finally set the new root and get it in the array
    nt->NodeArray[na_index] = new_root;
    nt->root = new_root;

    // DEBUGGING
    printf("NR DEBUG: Outgroup WAS adjusted\n");
  }

  // DEBUGGING
  //printf("NR DEBUG: Outgroup adjust routine finished\n");

  return;
}
