/*****************************************************************************/
/*                                                                           */
/*  RANDIP kappa version                                                     */
/*                                                                           */
/*  Software to perform random tree reconstruction for use with NR code.     */
/*                                                                           */
/*  Copyright (C) 2007 Monique M. Morin                                      */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify it  */
/*  under the terms of the GNU General Public License as published by the    */
/*  Free Software Foundation; either version 2 of the License, or (at your   */
/*  option) any later version.                                               */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful, but      */
/*  WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Public */
/*  License for more details.                                                */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License along  */
/*  with this program; if not, write to the Free Software Foundation, Inc.,  */
/*  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                    */
/*                                                                           */
/*  Monique Morin (morin@cs.unm.edu)                                         */
/*                                                                           */
/*  Monique Morin                                                            */
/*  University of New Mexico                                                 */
/*  Department of Computer Science                                           */
/*  Mail stop: MSC01 1130                                                    */
/*  1 University of New Mexico                                               */
/*  Albuquerque, NM 87131-0001                                               */
/*****************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<math.h>
#include<float.h>
#include<limits.h>
#include"globals.h"
#include"node.h"
#include"tree.h"
#include"twister.h"
#include"randrecon.h"


/******************************************************************************/
/*                            RandReconTree                                   */
/******************************************************************************/
void RandReconTree(TREE tr)
{
  NODE new_ancestor;
  NODE og;
  int index_a, index_b;

  // first set up some of the global stuff
  ResetAvailMarkersAll(tr, 1);

  available_count = num_extant_total - 1;  // do not count the outgroup
  og = tr->NodeArray[0];
  og->avail_marker = 0;   // do not let the outgroup be avail
  new_ancestor = NULL;

  while (available_count > 1)
  {
    index_a = FindAvailIndex(tr);
    tr->NodeArray[index_a]->avail_marker = 0;

    index_b = FindAvailIndex(tr);
    tr->NodeArray[index_b]->avail_marker = 0;

    new_ancestor = MakeSisters(tr, tr->NodeArray[index_a], tr->NodeArray[index_b]);
    new_ancestor->avail_marker = 1;

    available_count = available_count - 2 + 1;

    // DEBUGGING
    printf("RIP DEBUG:  node %d (internal id) is now the parent of %d/%d (left) and %d/%d (right)  \n", new_ancestor->internal_node_id, new_ancestor->left->internal_node_id, new_ancestor->left->external_node_id, new_ancestor->right->internal_node_id, new_ancestor->right->external_node_id);
    // END DEBUGGING
  }

  // make sure available count is now 1 so that we can stick 
  // the outgroup onto the structure and be done
  if (available_count == 1)
  {
    if (new_ancestor != NULL)
    {
      new_ancestor = MakeSisters(tr, new_ancestor, og);
    }
    else
    {
      // this was a case with only one extant and one outgroup so we call
      // MakeSisters differently
      if (num_extant_total == 2)
      {
        new_ancestor = MakeSisters(tr, tr->NodeArray[1], og);
      }
      else
      {
        printf("RIP ERROR:  Unexpected problem with Random Reconstruction -- program aborted\n");
	exit(0);
      }
    }
    tr->root = new_ancestor;
    tr->root->avail_marker = 0;

    available_count = available_count - 2 + 1;
  }
  else
  {
    printf("RIP ERROR:  Trouble adding in the outgroup at the last step of random reconstruction (available count is %d and it should be 1) -- program aborted\n", available_count);
    exit(0);
  }

  // one final check before we return
  if (available_count != 0)
  {
    printf("RIP ERROR:  Trouble random reconstruction -- program aborted\n");
    exit(0);
  }

  // DEBUGGING
  printf("RIP DEBUG:  our tree root is %d \n", tr->root->internal_node_id);

  return;
}

/******************************************************************************/
/*                           FindAvailIndex                                   */
/******************************************************************************/
int FindAvailIndex(TREE tr)
{
  int rand_val;
  int found, tries;
  NODE nd;
  NODE found_nd;

  tries = 0;
  found = 0;
  found_nd = NULL;

  while ( (found == 0) && (tries < max_tries) )
  {
    // choose from all the nodes in the tree and THEN see if
    // it is available -- do this so that although we will
    // have to repeat more often, all nodes have an equal
    // chance for becoming one of the two in the pair we
    // are going to make into sisters
    rand_val = (int) (genrand_real2() * tr->num_unique_nodes);

    nd = tr->NodeArray[rand_val];

    // DEBUGGING
    printf("RIP DEBUG:  rand val is %d\n", rand_val);

    // make sure it is avail and not the outgroup
    if ( (nd->outgroup == 0) && (nd->avail_marker == 1) )
    {
      found = 1;
      found_nd = nd;

      // DEBUGGING
      printf("RIP DEBUG:  node has been found for pairing its index is: %d (external id is %d)\n", found_nd->internal_node_id, found_nd->external_node_id);
      // END DEBUGGING
    }

    tries++;
  }


  // see if we exceeded max tries
  if (tries >= max_tries)
  {
    printf("RIP ERROR:  Max tries exceed while trying to find an available index -- program aborted\n");
    exit(0);
  }


  // if we made it here we are good to go
  return (found_nd->internal_node_id);

}

/******************************************************************************/
/*                             MakeSisters                                    */
/******************************************************************************/
NODE MakeSisters(TREE tr, NODE nd_a, NODE nd_b)
{
  NODE parent_nd;
  int rand_val;

  // DEBUGGING
  printf("starting make sisters routine\n");

  // in this routine we join the two given nodes by making a parent for
  // them and doing the appropriate hooking-up

  parent_nd = NewNode(tr->node_id_ctr);
  tr->NodeArray[tr->node_id_ctr] = parent_nd;
  tr->num_unique_nodes = tr->num_unique_nodes + 1;
  tr->node_id_ctr = tr->node_id_ctr++;

  // tell both nodes who their parent is now
  nd_a->parent = parent_nd;
  nd_b->parent = parent_nd;

  // now assign the parent the left and right kids
  rand_val = (int) (genrand_real2() * 2);

  if (rand_val == 0)
  {
    parent_nd->left = nd_a;   
    parent_nd->right = nd_b;

    // DEBUGGING
    printf("RIP DEBUG:  l/r was a/b\n");
  }
  else if (rand_val == 1)
  {
    parent_nd->left = nd_b;   
    parent_nd->right = nd_a;

    // DEBUGGING
    printf("RIP DEBUG:  l/r was b/a\n");
  }
  else
  {
    printf("RIP ERROR: Trouble with Make Sisters routine -- program aborted\n");
    exit(0);
  }

  // DEBUGGING
  printf("about to return new parent with internal id of %d\n", parent_nd->internal_node_id);

  return parent_nd;
}
